/*
 *  domain-wall-v3.cc
 *
 * started 2/11/94 by gp
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "dwm-v1.h"

#include "vector-1.h"
#include "wall-point-v3.h"
#include "domain-wall-v3.h"

extern struct MaterialParameters material_parameters;
extern struct RunParameters run_parameters;
extern Vector Xhat, Yhat, Zhat;

/* ===============  code for the DomainWall class ============== */

/* 
   This function creates a circular domain at a given position with
   a specified radius and number of points
*/

DomainWall::DomainWall( Vector Position, double radius, unsigned npoints, 
		       double initial_angle = 90.0)
{ 
  initialize(); // setup, for safety..,.

  // save the specified parameter as defaults.

  initial_x_position = Position.Get_Component(0); 
  initial_y_position = Position.Get_Component(1);
  initial_radius = radius; 
  initial_magnetization_angle = initial_angle;
  initial_n_wall_points = npoints; 


  Initialize_Bubble_Domain(Position, radius, npoints, initial_angle);
}


void DomainWall::Initialize_Bubble_Domain(Vector Position, double radius, unsigned npoints, double initial_angle = 90.0)
{
  int i;
  DWMfloat angle, delta_angle, magnetization_angle;
  Vector new_position(3), offset(3);
  WallPoint *wptr;

  remove_all_wall_points();         // first, kill the existing data structure...
  allocate_wall_points( npoints );  // .. and then allocate new points...


  // scale some of the physical quantities as necessary
  new_position = Position;
  new_position.Scale( 1.0 / material_parameters.distance_scale_factor );
  radius /= material_parameters.distance_scale_factor;

  // and set up some additional variables...
  
  delta_angle = TWOPI / ((double) npoints);
  initial_angle /= RADTODEG;

  for(i=0, angle = 0.0, wptr = wall_point_list; 
      i < npoints; 
      i++, angle += delta_angle, wptr = wptr->next) {

    offset.Set_Component(0, radius * cos(angle));
    offset.Set_Component(1, radius * sin(angle));

    magnetization_angle = Normalize_Angle(angle + initial_angle);

    wptr->Set_Position(new_position + offset, magnetization_angle); 
  }
}  

//void DomainWall::Initialize_Stripe_Domain()


void DomainWall::Initialize_Domain_From_File()
{
  FILE *fp;
  char line_buffer[4096]; int count;
  char word_buffer[128];
  char *temp;
  Vector offset(3);
  WallPoint *wptr;

  printf("initializing domain from file\n");


  if( (fp = fopen(initial_state_filename,"r")) == 0) {
    printf("Error initializing bubble.  I cannot find the file named %s\n", initial_state_filename);
    return;
  }
  
  int c;
  count = 0;
  char *buffer = line_buffer;
  while((c=fgetc(fp)) != EOF && c != '\n') *(buffer+count++) = c;
  *(buffer+count) = 0;

  //count = readline(line_buffer, fp);

  temp = line_buffer;
  temp = extract_token(temp, word_buffer);
  //printf("word_buffer = %s\n", word_buffer);

  temp = extract_token(temp, word_buffer);
  //printf("word_buffer = %s\n", word_buffer);

  temp = extract_token(temp, word_buffer);
  //printf("word_buffer = %s\n", word_buffer);
  
  unsigned npoints;
  sscanf(word_buffer,"%u", &npoints);

  remove_all_wall_points();         // first, kill the existing data structure...
  allocate_wall_points( npoints );  // .. and then allocate new points...

  double xtemp, ytemp, phitemp;

  int i;
  for(i=0, wptr = wall_point_list; 
      i < npoints; 
      i++, wptr = wptr->next) {

    temp = extract_token(temp, word_buffer);
    sscanf(word_buffer,"%lf", &xtemp);

    temp = extract_token(temp, word_buffer);
    sscanf(word_buffer,"%lf", &ytemp);

    temp = extract_token(temp, word_buffer);
    sscanf(word_buffer,"%lf", &phitemp);

    offset.Set_Component(0, xtemp / material_parameters.distance_scale_factor);
    offset.Set_Component(1, ytemp / material_parameters.distance_scale_factor);

    wptr->Set_Position(offset, phitemp);

    //printf("%f %f %f\n", xtemp, ytemp, phitemp);
  }
  fclose(fp);
  printf("done initializing domain from file.\n");

}  



void DomainWall::Reset()
{

  if(initial_state_filename != 0) {
    Initialize_Domain_From_File();
    return;
  }

  else {
    Vector new_position(3);
    new_position.Set_Component(0, initial_x_position);
    new_position.Set_Component(1, initial_y_position);
    new_position.Set_Component(2, 0.0);
    Initialize_Bubble_Domain(new_position, initial_radius, initial_n_wall_points, initial_magnetization_angle);
  }
}




void DomainWall::Examine_And_Correct_Shape()
{
  WallPoint *wp, *next_wp;
  double distance, curvature;
  int i;

/* Break this up into 2 loops -- one that looks th delete points, and
   another that adds points. */

/**************  
    if(distance < 7.0) {
      remove_wall_point(wp);
      printf("Removing a point...\n");
    }
*************/

  for(i=0,    wp = wall_point_list;      // Loop over wall points...
      i==0 || wp != wall_point_list; 
      i++,    wp = next_wp) { 

    next_wp = wp->next;

    distance = wp->delta_s_next;
    curvature = 0.5 * ( wp->curvature_1 + wp->next->curvature_1);

    if (curvature > 4.0e-2 && distance > 12.0) {
      add_wall_point(wp);
      if(run_parameters.debug) {
	printf("Adding point: distance and curvature.\n");
      }
    }

    else if (curvature > 2.0e-2 && distance > 18.0) {
      add_wall_point(wp);
      if(run_parameters.debug) {
	printf("Adding point: distance and curvature.\n");
      }
    }

    else if (distance > 28.0) {
      add_wall_point(wp);
      if(run_parameters.debug) {
	printf("Adding a point: distance only.\n");
      }
    }
  }
}


Vector DomainWall::Domain_Position()
{
  Vector domain_center(3);
  WallPoint *wptr = wall_point_list;

  if( wptr == 0 ) return(domain_center);

  /* average the position of all the domain wall points */
  int count = 1;
  domain_center = wptr->Position();

  for(wptr = wall_point_list->next; wptr!=wall_point_list; wptr = wptr->next) {
    domain_center = domain_center + wptr->Position();
    count++;
  }

  domain_center.Scale(1.0/((double) count));
  return(domain_center);
}

double DomainWall::Domain_Radius()
{
  WallPoint *wptr = wall_point_list;

  if( wptr == 0) return 0.0;  // there are no wall points allcated in the DW.

  Vector domain_center(3), local_radius(3);
  int count = 1;
  double radius;

  domain_center = wptr->Position();

  for(wptr = wall_point_list->next; wptr!=wall_point_list; wptr = wptr->next) {
    domain_center = domain_center + wptr->Position();
    count++;
  }

  domain_center.Scale(1.0/((double) count));

  wptr = wall_point_list;
  radius = (wptr->Position() - domain_center).Magnitude();

  for(wptr = wall_point_list->next; wptr!=wall_point_list; wptr = wptr->next) {
    radius += (wptr->Position() - domain_center).Magnitude();
  }

  return(radius/count);
}


double DomainWall::Domain_Length()
{
  WallPoint *wptr = wall_point_list;

  if( wptr == 0) return 0.0;  // there are no wall points allcated in the DW.

  Vector current_position(3), previous_position(3);
  double dlength = 0.0;

  previous_position = wall_point_list->Position();

  for(wptr = wall_point_list->next; wptr!=wall_point_list; wptr = wptr->next) {
    current_position = wptr->Position();
    dlength += (current_position - previous_position).Magnitude();
    previous_position = current_position;
  }

  /* now we need to do the last segment */

  current_position = wptr->Position();
  // previous_position is all set.
  dlength += (current_position - previous_position).Magnitude();

  /* ok: all done */

  return(dlength);
}


// this procedure calculates the length of the domain wall assuming
// that the variables delta_s_previous and delta_s_next are set 
// correctly in the WallPoint data objects.

double DomainWall::Domain_Length_From_Internals()
{
  WallPoint *wptr = wall_point_list;

  if( wptr == 0 ) return 0.0;  // there are no wall points allcated in the DW

  double dlength = 0.0;

  for(wptr = wall_point_list->next; wptr!=wall_point_list; wptr = wptr->next)
    dlength += wptr->delta_s_previous;

  // now do the last segment...

  dlength += wptr->delta_s_previous;

  return(dlength);
}


Vector  DomainWall::Maximum_Wall_Point_Displacement() 
{
  Vector maximum_displacement;
  WallPoint *wptr = wall_point_list;  // do the first by hand..

  maximum_displacement = wptr->delta_RSPhi;

  if(maximum_displacement.Get_Component(0) < 0.0) maximum_displacement.Scale_Component(0, -1.0);
  if(maximum_displacement.Get_Component(1) < 0.0) maximum_displacement.Scale_Component(1, -1.0);
  if(maximum_displacement.Get_Component(2) < 0.0) maximum_displacement.Scale_Component(2, -1.0);


  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next) {
    double dr, ds, dphi;

    dr = fabs(wptr->delta_RSPhi.Get_Component(0));
    ds = fabs(wptr->delta_RSPhi.Get_Component(1));
    dphi = fabs(wptr->delta_RSPhi.Get_Component(2));

    if(dr > maximum_displacement.Get_Component(0) ) maximum_displacement.Set_Component(0,dr);
    if(ds > maximum_displacement.Get_Component(1) ) maximum_displacement.Set_Component(1,ds);
    if(dphi > maximum_displacement.Get_Component(2) ) maximum_displacement.Set_Component(2,dphi);
  }
  return maximum_displacement;
}

void DomainWall::Step() 
{
  Examine_And_Correct_Shape();
  Update_Fields();
  Update_Internals();
  Calculate_Demag_Field();
  Calculate_Forces();
  Calculate_Delta_R();
  Calculate_Delta_S();
  Calculate_Delta_Phia();
  Update_Position();
}

void DomainWall::Update_Fields()
{
  if(wall_point_list == 0) return;  // there are no points to update!

  WallPoint *wptr = wall_point_list;  // do the first by hand..
  wptr->Update_Fields();  

  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
    wptr->Update_Fields();

}  


void DomainWall::Update_Internals()
{
  if(wall_point_list == 0) return;  // there are no points to update!

  WallPoint *wptr = wall_point_list;  //  do the first by hand..
  wptr->Update_Internals();  

  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
      wptr->Update_Internals();
}


void DomainWall::Calculate_Demag_Field()
{
  if(wall_point_list == 0) return;  // there are no points to update!

  WallPoint *wp1, *wp2;
  Vector Radius1(3), Radius2(3), Radius3(3);
  Vector my_wall_normal(3);
  Vector nhat1(3), nhat2(3),nhat3(3);
  Vector my_cross_product(3);
  double radius,dtheta,integrand,integrand_sign,integral;
  //double corrected_material_thickness;
  double source_material_thickness;
  double observation_material_thickness;
  int i,j;

  // THIS IS AN UGLY HACK TO determine the material thickness.
  for(i=0, wp1 = wall_point_list->next;
      wp1!=wall_point_list; 
      i++, wp1 = wp1->next) {
    source_material_thickness += Material_Thickness(wp1->Position());
  }
  source_material_thickness /= ((double)i);

  // here is the outer loop;
  for(i=0, wp1 = wall_point_list;
      i==0 || wp1!=wall_point_list; 
      i++, wp1 = wp1->next) {

    integral = 0.0;

    // here is the inner loop;

    for(j=0, wp2 = wp1->next; 
	j==0 || wp2!= wp1->next; 
	j++, wp2 = wp2->next) {
            
      Radius2 = wp2->Position() - wp1->Position();
      Radius1 = wp2->previous->Position() - wp1->Position();

      // calculate the corrected material thickness

      observation_material_thickness = Material_Thickness(wp1->Position());
//      source_material_thickness      = 0.5 * (Material_Thickness(wp2->Position()) +
//					      Material_Thickness(wp2->previous->Position()));
      //
      // BEGINNING OF THE CORRECTION TERMS...
      //

      Vector vtemp;
      double mydeltas = (wp2->Position() - wp2->previous->Position()).Magnitude();
      double mycurvature_1 = 0.5 *( wp2->curvature_1 + wp2->previous->curvature_1);

      Radius3 = wp2->Position() - wp2->previous->Position();
      my_wall_normal = CrossProduct(Radius3, Zhat);
      my_wall_normal.Normalize();

      Radius3 = wp2->Position() + wp2->previous->Position();
      Radius3.Scale(0.5);
      Radius3 = Radius3 - wp1->Position();

      // back to the usual...

      if( wp2->previous == wp1) {
	nhat1 = wp1->WallTangent;
	//nhat1 = wp1->Position() - wp1->previous->Position();
	nhat1.Normalize();
      }
      else {
	nhat1 = Radius1; nhat1.Normalize(); 
      }

      if(wp2 == wp1) { 
	nhat2 = wp1->WallTangent; nhat2.Scale(-1.0); 
	//nhat2 =  wp2->Position() - wp2->next->Position();
	nhat2.Normalize();
      }
      else { 
	nhat2 = Radius2; nhat2.Normalize(); 
      }

      // integrand sign...
      // this is to detect if we are looping back on ourselves.

      integrand_sign = 1.0;
      my_cross_product = CrossProduct(nhat1,nhat2);
      if(my_cross_product.Get_Component(2) < 0.0) { integrand_sign = -1.0; }

      double cos_theta = DotProduct(nhat1,nhat2);
      dtheta = acos(cos_theta);

      double rho2 = mydeltas * mydeltas * mycurvature_1 / 8.0;

      my_wall_normal.Scale(rho2);
      Radius3 = Radius3 + my_wall_normal;
      nhat3 = Radius3;
      nhat3.Normalize();
      
      // finally, calculate alpha1 and alpha2
      double cos_alpha1 = DotProduct(nhat3, nhat1);
      double alpha1 = acos(cos_alpha1);

      double cos_alpha2 = DotProduct(nhat3, nhat2);
      double alpha2 = acos(cos_alpha2);

      double radius1 = Radius1.Magnitude();
      double radius2 = Radius2.Magnitude();
      double radius3 = Radius3.Magnitude();

      double func = 
	0.5 * dtheta * (sqrt(pow(radius3, 2.0) + 
			     pow(source_material_thickness,2.0)) - 
			radius3 - 
			source_material_thickness);

      func += 
	0.5 * alpha1 * (sqrt(pow(radius1, 2.0) + 
			     pow(source_material_thickness,2.0)) - 
			radius1 - 
			source_material_thickness);

      func += 
	0.5 * alpha2 * (sqrt(pow(radius2, 2.0) + 
			     pow(source_material_thickness,2.0)) - 
			radius2 - 
			source_material_thickness);

      integrand = func;
      integral += integrand * integrand_sign;
    }
    wp1->Hdemag.Set_Component(2,
			      1.0 +  
			      (integral / (observation_material_thickness * 3.1415)));
  }
}

void DomainWall::Calculate_Forces()
{

  if(wall_point_list == 0) return;  // there are no points to update!

  WallPoint *wptr = wall_point_list;  //  do the first by hand..
  wptr->Calculate_Forces();  

  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
      wptr->Calculate_Forces();
}

/**********************************************************************
void DomainWall::Calculate_Delta_R()
{
  WallPoint *wp;
  double **lu_matrix, *lu_b, lu_d;
  int row;
  int *lu_indx;
  int npoints;
  int i,j;

  npoints = 2 * Number_Of_Wall_Points(); // this is the number of variables that we are solving for...

  lu_indx = (int *)malloc((npoints+1)*sizeof(int));

  lu_d = 1.0;
  lu_matrix = matrix(1,npoints,1,npoints);

  lu_b = vector(1,npoints);

  for(i=1; i <= npoints; i++) {
    for(j=1; j <= npoints; j++) {
      lu_matrix[i][j] = 0.0;
    }
    lu_b[i] = 0.0; lu_indx[i] = 0;
  }

  // if you build it, it will be solved...

  for(    i = 1, wp = wall_point_list; 
          i == 1 || wp!= wall_point_list;
          i++,   wp = wp->next) {

    row = 2*i - 1;

    int dr_index = row; 
    int dr_plus_index = row+2; 
    int dr_minus_index = row-2;

    int dphia_index = row+1; 
    int dphia_plus_index = row+3; 
    int dphia_minus_index = row-1;

    if(dr_plus_index > npoints) dr_plus_index -= npoints;
    if(dphia_plus_index > npoints) dphia_plus_index -= npoints;

    if(dr_minus_index < 1) dr_minus_index += npoints;
    if(dphia_minus_index < 1) dphia_minus_index += npoints;

    double previous_sin2_Phiw = sin(2.0 * wp->previous->Phiw);
    double next_sin2_Phiw     = sin(2.0 * wp->next->Phiw);
    double previous_cos2_Phiw = cos(2.0 * wp->previous->Phiw);
    double next_cos2_Phiw     = cos(2.0 * wp->next->Phiw);

    double delta_s_total = wp->delta_s_previous + wp->delta_s_next;

    double dSin2Phiw_dS =
      (wp->delta_s_previous / (delta_s_total * wp->delta_s_next)) * next_sin2_Phiw -
      (wp->delta_s_next / (delta_s_total * wp->delta_s_previous)) * previous_sin2_Phiw + 
      ((wp->delta_s_next - wp->delta_s_previous) / 
       (wp->delta_s_next * wp->delta_s_previous)) * sin(2.0 * wp->Phiw);

    double dCos2Phiw_dS =
      (wp->delta_s_previous / (delta_s_total * wp->delta_s_next)) * next_cos2_Phiw -
      (wp->delta_s_next / (delta_s_total * wp->delta_s_previous)) * previous_cos2_Phiw + 
      ((wp->delta_s_next - wp->delta_s_previous) / 
       (wp->delta_s_next * wp->delta_s_previous)) * cos(2.0 * wp->Phiw);

    double Cos2Phiw = cos(2.0 * wp->Phiw);
    double one_over_delta_S = 2.0 / delta_s_total;
    double my_delta_s = 0.5 * delta_s_total;

    DWMfloat dr_coefficient = 1.0 * my_delta_s * my_delta_s;
    DWMfloat dr_plus_coefficient = -0.5 * run_parameters.time_step * Cos2Phiw * my_delta_s;
    DWMfloat dr_minus_coefficient = +0.5 * run_parameters.time_step * Cos2Phiw * my_delta_s;

    DWMfloat dphia_coefficient = material_parameters.alpha * my_delta_s * my_delta_s - 
     run_parameters.time_step * Cos2Phiw * my_delta_s * my_delta_s +
     2.0 * material_parameters.quality_factor * run_parameters.time_step;

    DWMfloat dphia_plus_coefficient = - material_parameters.quality_factor * 
      run_parameters.time_step;

    DWMfloat dphia_minus_coefficient = - material_parameters.quality_factor * 
      run_parameters.time_step;

    // build the matrix.

    lu_matrix[row][dr_index]          = dr_coefficient;
    lu_matrix[row][dr_plus_index]     = dr_plus_coefficient;
    lu_matrix[row][dr_minus_index]    = dr_minus_coefficient;
    lu_matrix[row][dphia_index]       = dphia_coefficient;
    lu_matrix[row][dphia_plus_index]  = dphia_plus_coefficient;
    lu_matrix[row][dphia_minus_index] = dphia_minus_coefficient;
    lu_b[row] = wp->WallForce.Get_Component(1) * run_parameters.time_step * my_delta_s * my_delta_s;

    //
    // Ok. Now for the second equation...
    //

    dr_coefficient = material_parameters.alpha * my_delta_s * my_delta_s + 
      2.0 * run_parameters.time_step * (material_parameters.quality_factor - Cos2Phiw);

    dr_plus_coefficient = material_parameters.quality_factor - 0.5 * my_delta_s * dCos2Phiw_dS - Cos2Phiw;
    dr_plus_coefficient  *= -run_parameters.time_step;

    dr_minus_coefficient = material_parameters.quality_factor + 0.5 * my_delta_s * dCos2Phiw_dS - Cos2Phiw;
    dr_minus_coefficient *= - run_parameters.time_step;

    dphia_coefficient = (run_parameters.time_step * dCos2Phiw_dS - 1.0) * my_delta_s * my_delta_s;

    dphia_plus_coefficient = 0.5 * run_parameters.time_step * my_delta_s * Cos2Phiw;

    dphia_minus_coefficient = -0.5 * run_parameters.time_step * my_delta_s * Cos2Phiw;

    // build the matrix.
    row++;
    lu_matrix[row][dr_index]          = dr_coefficient;
    lu_matrix[row][dr_plus_index]     = dr_plus_coefficient;
    lu_matrix[row][dr_minus_index]    = dr_minus_coefficient;
    lu_matrix[row][dphia_index]       = dphia_coefficient;
    lu_matrix[row][dphia_plus_index]  = dphia_plus_coefficient;
    lu_matrix[row][dphia_minus_index] = dphia_minus_coefficient;
    lu_b[row] = (wp->WallForce.Get_Component(0) 
//		 - wp->curvature_1 * Cos2Phiw 
		 ) * run_parameters.time_step * my_delta_s * my_delta_s;
  }

  // done building the matrix.... now solve it

  ludcmp(lu_matrix, npoints, lu_indx, &lu_d);
  lubksb(lu_matrix, npoints, lu_indx, lu_b);

  // extract the results...store in dR_dt and dPhia_dt;

  for(i = 1,    wp = wall_point_list; 
      i == 1 || wp!= wall_point_list;
      i++,      wp = wp->next) {
    wp->delta_RSPhi.Set_Component(0, lu_b[2*i-1]);
    wp->dR_dt = wp->delta_RSPhi.Get_Component(0) / run_parameters.time_step;

    wp->delta_RSPhi.Set_Component(2, lu_b[2*i]);
    wp->dPhia_dt = wp->delta_RSPhi.Get_Component(2) / run_parameters.time_step;
  }

  free_matrix(lu_matrix,1,npoints,1,npoints);
  free_vector(lu_b,1,npoints);
  free(lu_indx);

}
**********************************************************************/


// This version allows for the non-uniform distribution of wall points.
// Installed 7/27/94

void DomainWall::Calculate_Delta_R()
{
  WallPoint *wp;
  double **lu_matrix, *lu_b, lu_d;
  int row;
  int *lu_indx;
  int npoints;
  int i,j;

  npoints = 2 * Number_Of_Wall_Points(); // this is the number of variables that we are solving for...

  lu_indx = (int *)malloc((npoints+1)*sizeof(int));

  lu_d = 1.0;
  lu_matrix = matrix(1,npoints,1,npoints);

  lu_b = vector(1,npoints);

  // zero out the matrix and associated vectors...
  for(i=1; i <= npoints; i++) {
    for(j=1; j <= npoints; j++) {
      lu_matrix[i][j] = 0.0;
    }
    lu_b[i] = 0.0; lu_indx[i] = 0;
  }

  // if you build it, it will be solved...

  for(    i = 1, wp = wall_point_list; 
          i == 1 || wp!= wall_point_list;
          i++,   wp = wp->next) {

    row = 2*i - 1;

    int dr_index = row; 
    int dr_plus_index = row+2; 
    int dr_minus_index = row-2;

    int dphia_index = row+1; 
    int dphia_plus_index = row+3; 
    int dphia_minus_index = row-1;

    if(dr_plus_index > npoints) dr_plus_index -= npoints;
    if(dphia_plus_index > npoints) dphia_plus_index -= npoints;

    if(dr_minus_index < 1) dr_minus_index += npoints;
    if(dphia_minus_index < 1) dphia_minus_index += npoints;

    double previous_cos2_Phiw = cos(2.0 * wp->previous->Phiw);
    double next_cos2_Phiw     = cos(2.0 * wp->next->Phiw);

    double delta_s_total = wp->delta_s_previous + wp->delta_s_next;

    double dCos2Phiw_dS =
      (wp->delta_s_previous / (delta_s_total * wp->delta_s_next)) * next_cos2_Phiw -
      (wp->delta_s_next / (delta_s_total * wp->delta_s_previous)) * previous_cos2_Phiw + 
      ((wp->delta_s_next - wp->delta_s_previous) / 
       (wp->delta_s_next * wp->delta_s_previous)) * cos(2.0 * wp->Phiw);

    double Cos2Phiw = cos(2.0 * wp->Phiw);


    DWMfloat timestep = run_parameters.time_step;

    DWMfloat dr_coefficient       = 1.0 - timestep * wp->d_1_0 * Cos2Phiw;
    DWMfloat dr_plus_coefficient  = - timestep * wp->d_1_plus  * Cos2Phiw;
    DWMfloat dr_minus_coefficient = - timestep * wp->d_1_minus * Cos2Phiw;

    DWMfloat dphia_coefficient = 
      material_parameters.alpha -
      timestep * 
	( Cos2Phiw + material_parameters.quality_factor * wp->d_2_0);

    DWMfloat dphia_plus_coefficient = 
      - timestep * material_parameters.quality_factor * wp->d_2_plus;

    DWMfloat dphia_minus_coefficient = 
      - timestep * material_parameters.quality_factor * wp->d_2_minus;

    // build the matrix.

    lu_matrix[row][dr_index]          = dr_coefficient;
    lu_matrix[row][dr_plus_index]     = dr_plus_coefficient;
    lu_matrix[row][dr_minus_index]    = dr_minus_coefficient;
    lu_matrix[row][dphia_index]       = dphia_coefficient;
    lu_matrix[row][dphia_plus_index]  = dphia_plus_coefficient;
    lu_matrix[row][dphia_minus_index] = dphia_minus_coefficient;
    lu_b[row] = wp->WallForce.Get_Component(1) * timestep;

    //
    // Now for the second equation...
    //

    dr_coefficient = 
      material_parameters.alpha + 
      timestep * 
	( Cos2Phiw * wp->d_2_0 + dCos2Phiw_dS * wp->d_1_0 -  
	 material_parameters.quality_factor * wp->d_2_0);

    dr_plus_coefficient = timestep *
      (Cos2Phiw * wp->d_2_plus + dCos2Phiw_dS * wp->d_1_plus - 
       material_parameters.quality_factor * wp->d_2_plus);

    dr_minus_coefficient = timestep *
      (Cos2Phiw * wp->d_2_minus + dCos2Phiw_dS * wp->d_1_minus - 
       material_parameters.quality_factor * wp->d_2_minus);

    dphia_coefficient = timestep * 
      ( Cos2Phiw * wp->d_1_0 + dCos2Phiw_dS) - 1.0;
    
    dphia_plus_coefficient = timestep * Cos2Phiw * wp->d_1_plus;

    dphia_minus_coefficient = timestep * Cos2Phiw * wp->d_1_minus;

    // build the matrix.
    row++;
    lu_matrix[row][dr_index]          = dr_coefficient;
    lu_matrix[row][dr_plus_index]     = dr_plus_coefficient;
    lu_matrix[row][dr_minus_index]    = dr_minus_coefficient;
    lu_matrix[row][dphia_index]       = dphia_coefficient;
    lu_matrix[row][dphia_plus_index]  = dphia_plus_coefficient;
    lu_matrix[row][dphia_minus_index] = dphia_minus_coefficient;
    lu_b[row] = wp->WallForce.Get_Component(0) * timestep;
  }

  // done building the matrix.... now solve it

  ludcmp(lu_matrix, npoints, lu_indx, &lu_d);
  lubksb(lu_matrix, npoints, lu_indx, lu_b);


  for(i = 1,    wp = wall_point_list; 
      i == 1 || wp!= wall_point_list;
      i++,      wp = wp->next) {
    wp->delta_RSPhi.Set_Component(0, lu_b[2*i-1]);

    /*****
      wp->dR_dt = wp->delta_RSPhi.Get_Component(0) / run_parameters.time_step;
      *****/

    wp->delta_RSPhi.Set_Component(2, lu_b[2*i]);

    /*******
      wp->dPhia_dt = wp->delta_RSPhi.Get_Component(2) / run_parameters.time_step;
      *******/
  }

  free_matrix(lu_matrix,1,npoints,1,npoints);
  free_vector(lu_b,1,npoints);
  free(lu_indx);
}



void DomainWall::Calculate_Delta_S()
{
  
  if(wall_point_list == 0) return;  // there are no points to update!

  double global_integral = 0.0;
  WallPoint *wp = wall_point_list;


  /* we could remove this function call since we are looping over the points below...*/
  double domain_length = Domain_Length();

  wp->normalized_distance = 0.0;
  wp->dS_dt = 0.0;
  
  for(wp = wall_point_list->next; wp!= wall_point_list; wp = wp->next) {
    wp->normalized_distance = 
      wp->previous->normalized_distance + wp->delta_s_previous;
  }

  for(wp = wall_point_list->next; wp!= wall_point_list; wp = wp->next) {
    wp->normalized_distance /= domain_length;
  }

  for(wp = wall_point_list->next; wp!= wall_point_list; wp = wp->next) {
    wp->dS_dt = (wp->previous)->dS_dt - 
      0.5 * ((wp->delta_RSPhi.Get_Component(0) * wp->curvature_1) + 
	     ((wp->previous)->delta_RSPhi.Get_Component(0) * (wp->previous)->curvature_1)) * 
	       (wp->normalized_distance - (wp->previous)->normalized_distance);
    
  }
  
  global_integral = wp->previous->dS_dt -  0.5 * 
    ((wp->delta_RSPhi.Get_Component(0) * wp->curvature_1) + 
     (wp->previous->delta_RSPhi.Get_Component(0) * wp->previous->curvature_1)) * 
       (1.0 - wp->previous->normalized_distance);

  global_integral *= -1.0;

  // now finish the job...

  wp = wall_point_list;
//  wp->dS_dt += (wp->normalized_distance * global_integral);
//  wp->dS_dt *= domain_length;
//  wp->delta_RSPhi.Set_Component(1, wp->dS_dt);

  for(wp=wall_point_list->next; wp!= wall_point_list; wp = wp->next) {
    wp->dS_dt = wp->dS_dt + wp->normalized_distance * global_integral;
    wp->dS_dt *= domain_length;
    wp->delta_RSPhi.Set_Component(1, wp->dS_dt);
  }
}


void DomainWall::Calculate_Delta_Phia()
{ 
  if(wall_point_list == 0) return;  // there are no points to update!

  WallPoint *wptr = wall_point_list;  //  do the first by hand..
  wptr->Calculate_Delta_Phia();  
  
  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
    wptr->Calculate_Delta_Phia();
}

void DomainWall::Update_Position()
{
  if(wall_point_list == 0) return;  // there are no points to update!

  WallPoint *wptr = wall_point_list;  //  do the first by hand..
  wptr->Update_Position();  

  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
      wptr->Update_Position();
}

void DomainWall::Print_Diagnostics()
{
  printf("====DomainWall diagnostics====\n");

  /* first, see if there are any points in the wall */

  if(wall_point_list == 0) {
    printf("No points in this domain wall.\n");
    printf("=============================\n");
    return;
  }

  /* now we definetly have some wall points to print, so print em */

  WallPoint *wp;
  wp = wall_point_list;

  wp->Print_Diagnostics();

  for(wp = wall_point_list->next; wp != wall_point_list; wp = wp->next) {
    wp->Print_Diagnostics();
  }

  printf("=============================\n");
}

void DomainWall::Print_Diagnostics_1()
{
  printf("========================DomainWall diagnostics========================\n");

  if(wall_point_list == 0) {
    printf("No points in this domain wall.\n");
    printf("=====================================================================\n");
    return;
  }

  /* now we definetly have some wall points to print, so print em */

  printf("%6s %10s %10s %10s %10s %10s %10s %10s\n",
	 "index",
	 "X    ",
	 "Y    ",
	 "Phia  ",
	 "Phiw  ",
	 "Phin  ",
	 "deltasnext",
	 "curvature");

  WallPoint *wp;  int i;
  for( i=0,  wp = wall_point_list;
       i == 0 || wp != wall_point_list;
       i++,      wp = wp->next) {
    printf("%6d %10.6lf %10.6lf %10.6lf %10.6lf %10.6lf %10.6lf %10.6lf\n",
	   i,
	   wp->Position().Get_Component(0),
	   wp->Position().Get_Component(1),
	   wp->Phia,
	   wp->Phiw,
	   wp->Phin,
	   wp->delta_s_next,
	   wp->curvature_1);
  }
  printf("=====================================================================\n");
}

void DomainWall::Print_Diagnostics_2()
{
  printf("========================DomainWall diagnostics========================\n");

  /* first, see if there are any points in the wall */

  if(wall_point_list == 0) {
    printf("No points in this domain wall.\n");
    printf("=====================================================================\n");
    return;
  }

  /* now we definetly have some wall points to print, so print em */

  printf("%6s %10s %10s %10s %10s %13s %13s %13s %13s %13s %13s\n",
	 "index",
	 "Phia  ",
	 "Phiw  ",
	 "Phin  ",
	 "deltasnext",
	 "curv  ",
	 "Fr    ",
	 "Fs    ",
	 "Dr    ",
	 "Ds    ",
	 "Dphia  ");


  WallPoint *wp;  int i;
  for( i=0,  wp = wall_point_list;
       i == 0 || wp != wall_point_list;
       i++,      wp = wp->next) {
    printf("%4d %10.6f %10.6f %10.6f %10.6f %10e %10e %10e %10e %10e %10e\n",
	   i,
	   wp->Phia,
	   wp->Phiw,
	   wp->Phin,
	   wp->delta_s_next,
	   wp->curvature_1,
	   wp->WallForce.Get_Component(0),
	   wp->WallForce.Get_Component(1),
	   wp->delta_RSPhi.Get_Component(0),
	   wp->delta_RSPhi.Get_Component(1),
	   wp->delta_RSPhi.Get_Component(2));
  }
  printf("=====================================================================\n");
}


void DomainWall::Print_Diagnostics_3()
{
  printf("========================DomainWall diagnostics========================\n");

  if(wall_point_list == 0) {
    printf("No points in this domain wall.\n");
    printf("=====================================================================\n");
    return;
  }

  /* now we definetly have some wall points to print, so print em */

  printf("%6s %10s %10s %10s %10s %10s %10s\n",
	 "index",
	 "d_1_+ ",
	 "d_1_0 ",
	 "d_1_- ",
	 "d_2_+ ",
	 "d_2_0  ",
	 "d_2_-  ");

  WallPoint *wp;  int i;
  for( i=0,  wp = wall_point_list;
       i == 0 || wp != wall_point_list;
       i++,      wp = wp->next) {
    printf("%6d %10.6lf %10.6lf %10.6lf %10.6lf %10.6lf %10.6lf\n",
	   i,
	   wp->d_1_plus,
	   wp->d_1_0,
	   wp->d_1_minus,
	   wp->d_2_plus,
	   wp->d_2_0,
	   wp->d_2_minus);
  }
  printf("=====================================================================\n");
}


void DomainWall::Print_Fields()
{
  printf("========================DomainWall Fields========================\n");

  /* first, see if there are any points in the wall */

  if(wall_point_list == 0) {
    printf("=====================================================================\n"); return;
  }

  /* now we definetly have some wall points to print, so print em */

  printf("%6s %10s %10s %10s %10s %10s %10s %10s\n",
	 "index",
	 "norm_dist",
	 "Hdemag   ",
	 "Hext     ",
	 "dS_dt    ",
	 "curv     ",
	 "ds_prev  ",
	 "ds_next  ");

  WallPoint *wp;  int i;
  for( i=0,  wp = wall_point_list;
       i == 0 || wp != wall_point_list;
       i++,      wp = wp->next) {
    printf("%4d %10.6f %10.6f %10e %10.6f %10.6f %10.6f %10.6f\n",
	   i,
	   wp->normalized_distance,
	   wp->Hdemag.Get_Component(2),
	   wp->Hexternal.Get_Component(2),
	   wp->dS_dt,
	   wp->curvature_1,
	   wp->delta_s_previous,
	   wp->delta_s_next);
  }
  printf("=====================================================================\n");
}


void DomainWall::Print_Point_Diagnostics(int point_offset, int index, FILE *fp)
{
  WallPoint *wp = wall_point_list;

  for(int i=0; i< point_offset; i++) wp = wp->next;

  fprintf(fp,"%d %f %f %f %f %f %f %f %f %f %f\n", index,
	  wp->Position().Get_Component(0),
	  wp->Position().Get_Component(1),
	  wp->Phia,
	  wp->Phiw,
	  wp->Phin,
	  wp->WallForce.Get_Component(0),
	  wp->WallForce.Get_Component(1),
	  wp->delta_RSPhi.Get_Component(0),
	  wp->delta_RSPhi.Get_Component(1),
	  wp->delta_RSPhi.Get_Component(2)
	  );
}

void DomainWall::Print_State_And_Phia(FILE *fp = NULL)
{
  WallPoint *wptr;
  FILE *myfp;

  if(fp==NULL) myfp = stdout;
  else myfp = fp;

  fprintf(myfp, "domain foo %d ", Number_Of_Wall_Points());

  wptr = wall_point_list;
  fprintf(myfp, "%f %f %f ", 
	  wptr->Position().Get_Component(0) * material_parameters.distance_scale_factor, 
	  wptr->Position().Get_Component(1) * material_parameters.distance_scale_factor,
	  wptr->Phia);

  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
    fprintf(myfp, "%f %f %f ", 
	    wptr->Position().Get_Component(0) * material_parameters.distance_scale_factor,
	    wptr->Position().Get_Component(1) * material_parameters.distance_scale_factor,
	    wptr->Phia);
}


void DomainWall::allocate_wall_points(unsigned number_of_points)
{
  WallPoint *current_wp, *previous_wp;

  /* allocate the first wall point by hand */

  wall_point_list = new WallPoint;
  current_wp = wall_point_list;

  /* now do the rest of the points */

  for(int i=1; i < number_of_points; i++) {
    
    previous_wp = current_wp;
    current_wp = new WallPoint;

    previous_wp->next = current_wp;
    current_wp->previous = previous_wp;

  }

  /* finally we join the two ends to make a loop! */

  current_wp->next = wall_point_list;
  wall_point_list->previous = current_wp;

  /* and set the number of wall points */
  n_wall_points = number_of_points;
}

//    
// this code adds a wall point between wp1 and wp1->next
//
void DomainWall::add_wall_point(WallPoint *wp1)
{
  if ( wall_point_list == 0 ) {
    printf("Error: you have tried to interpolate a new point between\n");
    printf("       two points, and you dont have any points!\n");
    exit(0);
  }

  WallPoint *new_wp= new WallPoint;

  new_wp->previous = wp1;
  new_wp->next     = wp1->next;

  wp1->next->previous = new_wp;
  wp1->next = new_wp;
  n_wall_points++;

  // Now we setup the variables that we need...
  //   1. the new point position is the average of the two adjacent positions..

  new_wp->Position() = (new_wp->previous->Position() + new_wp->next->Position());
  new_wp->Position().Scale(0.5);

  Vector tangent, normal, vtemp;
  tangent = new_wp->previous->Position() - new_wp->next->Position();
  double mydeltas = tangent.Magnitude();
  tangent.Normalize();
  normal  = CrossProduct( tangent, Zhat );

  vtemp = new_wp->previous->WallTangent - new_wp->next->WallTangent;

  double mycurvature_1 = vtemp.Magnitude() / mydeltas;

  normal.Scale( mydeltas * mydeltas * mycurvature_1 / 8.0);

  new_wp->Position() =  new_wp->Position() - normal;

  //   2. the new angle is the average of the two adjacent angles...
  //   BEWARE OF THE BRANCH CUT...

  new_wp->Phia = new_wp->previous->Phia +
    Normalize_Angle(new_wp->next->Phia, new_wp->previous->Phia);
  new_wp->Phia *= 0.5;

  new_wp->Phia = Normalize_Angle(new_wp->Phia);

  // All done.... I think.

}

void DomainWall::add_wall_point(int offset)
{
  if ( wall_point_list == 0 ) {
    printf("Error: you have tried to interpolate a new point between\n");
    printf("       two points, and you dont have any points!\n");
    exit(0);
  }

  WallPoint *wp= wall_point_list;

  for(int i=0; i<offset; i++) {
    wp = wp->next;
  }

  add_wall_point(wp);
}



//
// this code removes the wall point wp1.
//
void DomainWall::remove_wall_point(WallPoint *wp1)
{
  if ( wall_point_list == 0 ) return;

  if(wp1 == 0) {
    printf("Warning: you have tried to remove a NULL wall point\n");
  }

  // dont loose the pointer to the list!!!

  if(wp1 == wall_point_list) { 
    wall_point_list = wp1->previous->previous;
  }

  wp1->previous->next = wp1->next;
  wp1->next->previous = wp1->previous;

  delete wp1;
  n_wall_points--;
}

void DomainWall::remove_all_wall_points()
{
  
  if(wall_point_list == 0) return;  // we are already done!

  WallPoint *w, *next_to_kill;

  for( w=wall_point_list->previous ; w != wall_point_list; w = next_to_kill )
    {
      next_to_kill = w->previous;
      delete w;
    }

  // dont forget to kill the last point....and set the pointer to null.

  delete wall_point_list;
  wall_point_list = 0;
  n_wall_points = 0;
}

  

void DomainWall::Calculate_Normalized_Distance()
{
  if(wall_point_list == 0) return;  // there are no points to update!
  WallPoint *wp = wall_point_list;

  double domain_length = Domain_Length();

  wp->normalized_distance = 0.0;

  for(wp = wall_point_list->next; wp!= wall_point_list; wp = wp->next) {
    wp->normalized_distance = 
      wp->previous->normalized_distance + 
	(wp->delta_s_previous / domain_length);
  }
}


void DomainWall::Write(FILE *fp)
{
  WallPoint *wptr;
  FILE *myfp;

  if(fp==NULL) myfp = stdout;
  else myfp = fp;

  fprintf(myfp, "domain foo %d ", Number_Of_Wall_Points());

  wptr = wall_point_list;
  fprintf(myfp, "%f %f %f ", 
	  wptr->Position().Get_Component(0) * material_parameters.distance_scale_factor, 
	  wptr->Position().Get_Component(1) * material_parameters.distance_scale_factor,
	  wptr->Phia);

  for(wptr = wall_point_list->next; wptr != wall_point_list; wptr = wptr->next)
    fprintf(myfp, "%f %f %f ", 
	    wptr->Position().Get_Component(0) * material_parameters.distance_scale_factor,
	    wptr->Position().Get_Component(1) * material_parameters.distance_scale_factor,
	    wptr->Phia);

}
