#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ctype.h>

#include "dwm-dfield.h"
#include "dwm-v1.h"

int readline(char *, FILE *);

void DWMDiscreteField::initialize_with_defaults()
{
  lattice_constant_x = 0.0;
  lattice_constant_y = 0.0;
  lattice_constant_product = 0.0;

  lattice_origin_x = 0.0;
  lattice_origin_y = 0.0;
  lattice_extent_x = 0.0;
  lattice_extent_y = 0.0;

  number_of_cells_x = 0;
  number_of_cells_y = 0;
  number_of_components = 0;
  
  description = 0;
  source = 0;
  pattern = 0;
  data = 0;
}

void DWMDiscreteField::allocate_storage_space()
{
  number_of_components = 1;
  unsigned nelements = number_of_cells_x * number_of_cells_y *  number_of_components;

  data = new double[nelements];
  for(int i = 0; i < nelements; i++) *(data+i) = 0.0;
}

void DWMDiscreteField::allocate_storage_space(unsigned nx, unsigned ny, 
				      unsigned ncomponents) {
  number_of_cells_x = nx;
  number_of_cells_y = ny;
  number_of_components = ncomponents;

  assert( ncomponents == 1 || ncomponents == 3);

  unsigned nelements = nx * ny * ncomponents;

  data = new double[nelements];
  for(int i = 0; i < nelements; i++) *(data+i) = 0.0;

}

void DWMDiscreteField::remove_storage_space()
{
  delete [] data;
  data = 0;
}


void DWMDiscreteField::reset()
{
  number_of_cells_x = 0;
  number_of_cells_y = 0;
  number_of_components = 0;

  lattice_constant_product = 0.0;
  lattice_constant_x = 0.0;
  lattice_constant_y = 0.0;

  lattice_origin_x = 0.0;
  lattice_origin_y = 0.0;
  lattice_extent_x = 0.0;
  lattice_extent_y = 0.0;

  delete [] data;
  delete [] source;
  delete [] pattern;
  delete [] description;

  data = 0; source = 0;
  pattern = 0; description = 0;
}

double& DWMDiscreteField::map_node(unsigned x, unsigned y, unsigned nc)
{
  //printf("%d\t%d\n",x,y);
  assert(x >= 0 && x < number_of_cells_x);
  assert(y >= 0 && y < number_of_cells_y);

  return *(data + y * number_of_cells_x + x );
}


double DWMDiscreteField::Interpolate(double x, double y, unsigned component)
{
  double a1, a2, a3, a4, result;

  // First check to see if we are outside the region that is
  // described by the mesh, and do something reasonable if this is
  // the case...

  if(x < lattice_origin_x || x >= lattice_extent_x) return 0.0;
  if(y < lattice_origin_y || y >= lattice_extent_y) return 0.0;

  // Get the mesh coordinates of the data point that is to the lower left 
  // of the specified coordinates.

  unsigned xint = (int)((x - lattice_origin_x) / lattice_constant_x);
  unsigned yint = (int)((y - lattice_origin_y) / lattice_constant_y);

  // x1 and y1 are the physical coordinates of the point described above.
  // They are used below in the interpolation algorithm.

  double x1 = xint *  lattice_constant_x;
  double y1 = yint *  lattice_constant_y;

  // Lookup the four adjacent mesh points...
  
  a1 = map_node(xint,   yint,   component);
  a2 = map_node(xint+1, yint,   component);
  a3 = map_node(xint,   yint+1, component);
  a4 = map_node(xint+1, yint+1, component);

  // ...and finally do the interpolation.
  
  result = a1 * ( 1.0 - (x-x1)/lattice_constant_x - (y-y1)/ lattice_constant_y +
		 (x-x1)*(y-y1)/(lattice_constant_x * lattice_constant_y) );

  result += a2 * ((x-x1)/ lattice_constant_x - 
		  (x-x1)*(y-y1)/(lattice_constant_x * lattice_constant_y));

  result += a3 * ((y-y1)/ lattice_constant_y - 
		  (x-x1)*(y-y1)/(lattice_constant_x * lattice_constant_y));

  result += a4 * (x-x1) * (y-y1) / ( lattice_constant_x * lattice_constant_y);

  //printf("Interpolate: result = %f\n", result);

  return result;
}

int readline(char *buffer, FILE *fp)
{
  int c,i=0;
  while((c=fgetc(fp)) != EOF && c != '\n') *(buffer+i++) = c;
  *(buffer+i) = 0;
  return i;
}

// This procedure reads a field profile from an ascii file assuming that all of the 
// internal parameters have been set.

/***************
void DWMDiscreteField::Read_From_Ascii_File(char *filename)
{
  int k;  double dtemp; char *ptr;
  char line_buffer[4096], word_buffer[64];
  FILE *fp;

  // First check that the internals make sense...
  if( number_of_cells_x < 1 || number_of_cells_y < 1) {
    printf("Error: the number of cells has not been set.\n"); return; 
  }

  if(lattice_constant_x == 0.0 || lattice_constant_y == 0.0) {
    printf("Error: the lattice constants have not been set.\n"); return; 
  }

  remove_storage_space();
  allocate_storage_space(); // according to the internals...

  if((fp=fopen(filename,"r")) == NULL) {
    printf("Error opening file name %s\n", filename);
    printf("Data not loaded into DWMDiscreteField\n");
    return;
  }
  
  // allocate storage for the source filename and save it...
  source = new char[ (strlen(filename)+1) * sizeof(char) ];
  strcpy( source, filename );

  printf("Reading magnetic field from file %s....", filename);

  for(int j=0; j < number_of_cells_y; j++) {

    // read in a line of data...
    if(readline(line_buffer, fp) == 0) {
      printf("Error: Not enough data in DWMDiscreteField ASCII File.\n");
      return;
    }

    // extract the data from the string in line_buffer
    ptr = line_buffer;

    for(int i=0; i < number_of_cells_x; i++) {

      k=0;
      while(*ptr != ' ' && *ptr != '\n') word_buffer[k++] = *ptr++;
      word_buffer[k] = 0;
      ptr++;
      if(sscanf(word_buffer,"%lf",&dtemp) != 1) {
	printf("Error scanning argument: %s while reading ascii field file. Result = %f\n",
	       word_buffer, dtemp);
      }
      Set_Node(i,j,1, dtemp);
    }
  }
  fclose(fp);
  printf("done\n");
  recompute_internals();
}
***************/


void DWMDiscreteField::Read_From_Ascii_File(char *filename)
{
  int k;  double dtemp; char *ptr;
  char line_buffer[4096], word_buffer[64];
  FILE *fp;

  if((fp=fopen(filename,"r")) == NULL) {
    printf("Error opening file name %s\n", filename);
    return;
  }
  
  // allocate storage for the source filename and save it...
  source = new char[ (strlen(filename)+1) * sizeof(char) ];
  strcpy( source, filename );

  printf("Reading magnetic field from file %s....", filename);

  // extract header information.....

  while(readline(line_buffer, fp) != 0) {
    char *p;
    p=strip_white(line_buffer);

    if( *p == '#' ) continue;  // lines that begin with # are comments.....

    if(!isalpha((int)*p)) break;  // this is data... dont mess with it here

    // here we have a command string... process it.
    
    while((p=extract_token(p, word_buffer)) != 0) {

      if(!strcmp(word_buffer,"format")) {
	p=extract_token(p, word_buffer);
	// strip off the initial keyword and then save the formatting info...
      }
      
      if(!strcmp(word_buffer,"pattern")) {
	//p=extract_token(p, word_buffer);
	// strip off the initial keyword and then save the pattern data.
	pattern = new char[ (strlen(p)+1) * sizeof(char) ];
	strcpy(pattern, p);
	//printf("Pattern: %s\n", p);
	// also, get out of this while loop as we need to get to the next line
	break;
      }
      
      if(!strcmp(word_buffer,"dx")) {
	p=extract_token(p, word_buffer);
	double dx_temp;
	sscanf(word_buffer,"%lf", &dx_temp);
	Set_Lattice_Constant(0U,dx_temp);
      }

      if(!strcmp(word_buffer,"dy")) {
	p=extract_token(p, word_buffer);
	double temp;
	sscanf(word_buffer,"%lf", &temp);
	Set_Lattice_Constant(1U,temp);
      }

      if(!strcmp(word_buffer,"xsize")) {
	p=extract_token(p, word_buffer);
	int temp;
	sscanf(word_buffer,"%d", &temp);
	//printf("Setting lattice size x to %d\n", temp);
	Set_Lattice_Size(0,temp);
      }

      if(!strcmp(word_buffer,"ysize")) {
	p=extract_token(p, word_buffer);
	int temp;
	sscanf(word_buffer,"%d", &temp);
	//printf("Setting lattice size y to %d\n", temp);
	Set_Lattice_Size(1,temp);
      }
    }
  }

  // now extract the field profile
  // check that the internals make sense...
  if( number_of_cells_x < 1 || number_of_cells_y < 1) {
    printf("Error: the number of cells has not been set.\n"); return; 
  }

  if(lattice_constant_x == 0.0 || lattice_constant_y == 0.0) {
    printf("Error: the lattice constants have not been set.\n"); return; 
  }

  //remove_storage_space();
  allocate_storage_space(); // according to the internals...
  
  // assume that a line has already been read and is in place...

  for(int j=0; j < number_of_cells_y; j++) {

    // a line has already been read above...but dont forget to get another
    // at the bottom of the loop.

    if(j!=0) {
      if(readline(line_buffer, fp) == 0) {
	printf("Error: Not enough data in DWMDiscreteField ASCII File.\n");
	return;
      }
    }

    // extract the data from the string in line_buffer
    ptr = line_buffer;

    for(int i=0; i < number_of_cells_x; i++) {

      k=0;
      while(*ptr != ' ' && *ptr != '\n') word_buffer[k++] = *ptr++;
      word_buffer[k] = 0;
      ptr++;
      if(sscanf(word_buffer,"%lf",&dtemp) != 1) {
	printf("Error scanning argument: %s while reading ascii field file. Result = %f\n",
	       word_buffer, dtemp);
      }
      Set_Node(i,j,1, dtemp);
    }

  }
 
  fclose(fp);
  printf("done\n");
  recompute_internals();
}

void DWMDiscreteField::Print_Diagnostics()
{
  printf("No. of cells along x: %d\n", number_of_cells_x);
  printf("No. of cells along y: %d\n", number_of_cells_y);
  printf("No. of components:    %d\n", number_of_components);

  printf("X Lattice constant:       %f\n", lattice_constant_x);
  printf("Y Lattice constant:       %f\n", lattice_constant_y);
  printf("Lattice constant product: %f\n", lattice_constant_product);

  printf("X lattice origin:       %f\n", lattice_origin_x);
  printf("Y lattice origin:       %f\n", lattice_origin_y);

  printf("X lattice extent:       %f\n", lattice_extent_x);
  printf("Y lattice extent:       %f\n", lattice_extent_y);

  printf("Description: %s\n", description);
  printf("Source: %s\n", source);
}


void DWMDiscreteField::Print_Data_Field()
{

  for(int j=0; j< number_of_cells_y; j++) {
      for(int i=0; i< number_of_cells_x; i++) {
	printf("%3.1f  ", *(data + j * number_of_cells_x + i ));
      }
      printf("\n");
    }
}


void DWMDiscreteField::recompute_internals()
{
  lattice_constant_product = lattice_constant_x * lattice_constant_y;

  lattice_extent_x = lattice_origin_x + 
    lattice_constant_x * (number_of_cells_x - 1);

  lattice_extent_y = lattice_origin_y + 
    lattice_constant_y * (number_of_cells_y - 1);
}


void DWMDiscreteField::Write_Object_Description(FILE *fp)
{
  char *p;
  char wbuffer1[32], wbuffer2[32];
  double xc, yc;

  p = pattern;

  fprintf(fp,"pattern ");

  do {
    p = extract_token(p, wbuffer1);
    p = extract_token(p, wbuffer2);

    sscanf(wbuffer1,"%lf", &xc);
    sscanf(wbuffer2,"%lf", &yc);

    //printf("%s %s %lf %lf \n", wbuffer1, wbuffer2, xc, yc);

    fprintf(fp, "%lf %lf ", xc  + lattice_origin_x, yc + lattice_origin_y);
  }  while( *p != 0);
    fprintf(fp,"\n");
}


