//
// dwm-dfield.h
//
// This header file provides the definition for the DWMDiscreteField object.
// This object is derived from the abstract class DWMField, which is the 
// prototype object for all field objects.
//
// 4/20/94 started by gp
// 8/24/94 renamed to dwm-dfield.h.  Made this object one derived from 
// DWMField

#ifndef _DWM_DFIELD_H
#define _DWM_DFIELD_H

#include "dwm-tfield.h"

class DWMDiscreteField : public DWMTField {

public:
  DWMDiscreteField() { initialize_with_defaults(); }
  DWMDiscreteField(int nx, int ny, int components) { 
    initialize_with_defaults(); 
    allocate_storage_space(nx, ny, components); 
  }

  ~DWMDiscreteField() { remove_storage_space(); }

  // here are the functions that we must provide to make this
  // a concrete class.  The following are required by the ultimate base
  // base object DWMObject...

  virtual DWMClassType Class_Type() const { return DWMDiscreteFieldClass; }
  virtual char *Class_Name() const { return "DWMDiscreteFieldClass"; }

  // ... and this function must be defined per the requirement in DWMField. 

  virtual double Evaluate_Field(double x, double y, double t)
    { 
      double fresult = Interpolate(x,y,0);
      double tresult = TScale(t);
      return(fresult * tresult); }

  // ok. now for the user interface routines...

  void Resize(int nx, int ny, int components) { 
    remove_storage_space();
    allocate_storage_space(nx, ny, components); 
  }

  //Read_From_File(char *filename);
  void Read_From_Ascii_File(char *filename);
  void Read_From_Ascii_File(char *filename, unsigned, unsigned, double, double);
  //Read_From_Binary_File(char *filename);
  //Write_To_File(char *filename);

  double Interpolate(double x, double y, unsigned component);

  void Set_Node(unsigned x, unsigned y, unsigned component, double value)
    { map_node(x,y,component) = value; }

  double Node(unsigned x, unsigned y, unsigned component)
    { return map_node(x,y,component); }

  /* Code to get and set the lattice constant associated with the 
   X and Y axes respectively...*/


  // 3/16/95 -- recoded a few procedures below so that I use only
  // one return statement.  This is apparently a problem with the
  // SGI C++ compiler.

  double Lattice_Constant(unsigned axis) {
    double result;
    assert(axis == 0 || axis == 1);
    if(axis == 0) result = lattice_constant_x;
    if(axis == 1) result = lattice_constant_y;
    return result;
  }

  void Set_Lattice_Constant(unsigned axis, double value) { 
    assert(axis == 0 || axis == 1);
    if(axis == 0) lattice_constant_x = value; 
    if(axis == 1) lattice_constant_y = value;
    recompute_internals();
  }

  void Set_Lattice_Constant(double x_value, double y_value) { 
    lattice_constant_x = x_value; 
    lattice_constant_y = y_value;
    recompute_internals();
  }

  /* Code to set and get the origin of the lattice */

  // 3/18/95 -- recoded for the SGI compiler. gnp.
  double Lattice_Origin(unsigned axis) {
    double result;
    assert(axis == 0 || axis == 1);
    if(axis == 0) result = lattice_origin_x;
    if(axis == 1) result = lattice_origin_y;
    return result;
  }

  void Set_Lattice_Origin(unsigned axis, double value) { 
    assert(axis == 0 || axis == 1);
    if(axis == 0) lattice_origin_x = value; 
    if(axis == 1) lattice_origin_y = value;
    recompute_internals();
  }

  void Set_Lattice_Origin(double x_value, double y_value) { 
    lattice_origin_x = x_value; 
    lattice_origin_y = y_value;
    recompute_internals();
  }

  /* Code to get the upper limits of the mesh along X and Y axes...*/

  double Lattice_Extent(unsigned axis) {
    double result;
    assert(axis == 0 || axis == 1);
    if(axis == 0) result=lattice_extent_x;
    if(axis == 1) result=lattice_extent_y;
    return result;
  }


  /* Code to set/get the size of the lattice.  Intended to be used to
     do a read....*/

  unsigned Lattice_Size(unsigned axis) {
    unsigned result;
    assert(axis == 0 || axis == 1);
    if(axis == 0) result=number_of_cells_x;
    if(axis == 1) result=number_of_cells_y;
    return result;
  }

  void Set_Lattice_Size(unsigned axis, unsigned value) { 
    assert(axis == 0 || axis == 1);
    if(axis == 0) number_of_cells_x = value; 
    if(axis == 1) number_of_cells_y = value;
  }

  // and of course some diagnostic code...
  void Print_Diagnostics();
  void Print_Data_Field();

  // finally, we use this to write out the pattern...
  void Write_Object_Description(FILE *fp);


private:

  void initialize_with_defaults();
  void reset();
  void allocate_storage_space();
  void allocate_storage_space(unsigned nx, unsigned ny, unsigned ncomponents);
  void remove_storage_space();
  void recompute_internals();
  double& map_node(unsigned x, unsigned y, unsigned nc);

  // Lattice constants (and their product) for the 
  // regular array of numbers that is stored in data.

  double lattice_constant_x;
  double lattice_constant_y;
  double lattice_constant_product;

  // Origin (x and y) and extent of the mesh.
  double lattice_origin_x;
  double lattice_origin_y;
  double lattice_extent_x;
  double lattice_extent_y;

  // Number of cells and components..
  unsigned number_of_cells_x;
  unsigned number_of_cells_y;
  unsigned number_of_components;

  // Other, possibly relevant information..
  char *description;
  char *source;

  // store the associated pattern info.  for now lets keep it simple
  char *pattern;

  // Finally the data...
  double *data;

};

#endif
