/*
 * dwm-field-utilities.cc
 *
 * Started: Nov 14, 1994
 * Author : George Patterson
 *
 * Description: This file contains several utility procedures that are used
 * to evaluate the magnetic field produced by a groove or ridge change in the
 * material height.  All of these procedures are low level, and are called by
 * the code that initialized the field profile in the groove field data 
 * structutes.
 *
 *
 *
 * Change Log:
 * 12/19/94 -- modified groove_component_z to handel the case when the
   argument z3 is equal to 0.0 . In this case, the result should either
   be Pi/2 or -Pi/2 depending on the sign of z1 and z2.

 * 1/20/95 -- added code to correct dipole formula for delta function
   located at the origin with strength 8pi/3

 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "dwm-v1.h"

/* utility procedure called by higher level functions. 
 * Part of the response function of cube of magnetic material.
 */
double groove_component_z(double z1, double z2, double z3)
{
  double result;

  if(z3 == 0.0) {
    if( z1 * z2 > 0.0 ) return(-1.0 * PIOVERTWO);
    if( z1 * z2 < 0.0 ) return(PIOVERTWO);
    printf("Error: the argument to atan cannot be intepreted. z1 = %f, z2 = %f, z3 = %f\n", z1, z2, z3);
    exit(0);
  }

  result = -1.0 * atan( (z1 * z2) / (z3 * sqrt( z1*z1 + z2*z2 + z3*z3 )) );
  
  return result;
}

double groove_field_z(double x, double y, double z, double lx, double ly, double lz)
{
  double temp = 0.0;

  temp += groove_component_z( x + lx, y + ly, z + lz);
  temp += groove_component_z( x - lx, y - ly, z + lz);
  temp += groove_component_z( x + lx, y - ly, z - lz);
  temp += groove_component_z( x - lx, y + ly, z - lz);

  temp -= groove_component_z( x - lx, y - ly, z - lz);
  temp -= groove_component_z( x - lx, y + ly, z + lz);
  temp -= groove_component_z( x + lx, y - ly, z + lz);
  temp -= groove_component_z( x + lx, y + ly, z - lz);


//  if(x-lx <= 0.0 && x+lx >= 0.0 &&
//     y-ly <= 0.0 && y+ly >= 0.0 &&
//     z-lz <= 0.0 && z+lz >= 0.0) temp += 8.377580;

  if(x-lx < 0.0 && x+lx > 0.0 &&
     y-ly < 0.0 && y+ly > 0.0 &&
     z-lz < 0.0 && z+lz > 0.0) temp += 8.377580;

  return(temp);
}

double average_groove_field_z(double x, double y,
			      double base_material_height,
			      double effective_material_height, 
			      double lx, double ly, double lz)
{
  unsigned npoints = 5;
  double step_size = effective_material_height / ((double) npoints-1);
  double integral = 0.0;
  double z = 0.0;
  double temp = 0.0;

  double lower_integration_limit = lz - base_material_height;
  double upper_integration_limit = lower_integration_limit + effective_material_height;

  integral  = groove_field_z(x,y,lower_integration_limit,lx,ly,lz);
  integral += groove_field_z(x,y,upper_integration_limit - 0.01 ,lx,ly,lz);
  
  unsigned i;
  for(i=1, z=lower_integration_limit + step_size; 
      i< npoints-1; 
      i++, z+=step_size) {
    temp = groove_field_z(x,y,z,lx,ly,lz);

    if(i%2 == 0) 
      integral += 2.0 * temp;
    else 
      integral += 4.0 * temp;
  }

  integral /= (3.0 * ((double) npoints-1));

  return integral;

}

// 1/24/95 -- modified ridge code to deal with the region directly under the ridge in a
// more detailed manner....

double average_ridge_field_z(double x, double y,
			      double base_material_height,
			      double effective_material_height, 
			      double lx, double ly, double lz)
{
  unsigned npoints = 11;
  double integral_1 = 0.0;
  double integral_2 = 0.0;
  double step_size;
  double z = 0.0;
  double temp = 0.0;
  double lower_integration_limit;
  double upper_integration_limit;


  if( x >= -lx && x <= lx && y >= -ly && y <= ly ) {
    //
    // here we are directly under the ridge. Break the integral into two parts.
    // (Note: this decomposition assumes that the grooves/ridges are not overlayed!!!)

    // First part..... average the field produced by the ridge over the original material.
    
    lower_integration_limit = -1.0 * lz - base_material_height;
    upper_integration_limit = -1.0 * lz;

    // compute the step size...
    step_size = (upper_integration_limit - lower_integration_limit) / ((double) npoints-1);

    // do the endpoints by hand because they get different scaling factors and this is where
    // we may displace the function evaluation to avoid numerical problems.

    integral_1 += groove_field_z(x, y, lower_integration_limit,        lx, ly, lz);
    integral_1 += groove_field_z(x, y, upper_integration_limit - 0.01 ,lx, ly, lz);  

    // now get the remainder of the terms in the integral...
  
    unsigned i;
    for(i=1, z=lower_integration_limit + step_size; 
	i< npoints-1; 
	i++, z+=step_size) {
      temp = groove_field_z(x,y,z,lx,ly,lz);

      if(i%2 == 0) 
	integral_1 += 2.0 * temp;
      else 
	integral_1 += 4.0 * temp;
    }
    
    //integral_1 /= (3.0 * ((double) npoints-1));
    integral_1 *= step_size / 3.0;
    

    //
    // Second part..... average the field produced by the ridge over itself.
    //

    lower_integration_limit = -1.0 * lz;
    upper_integration_limit =  lz;

    // compute the step size...
    step_size = (upper_integration_limit - lower_integration_limit) / ((double) npoints-1);

    // displace the evaluation of function at the upper point to avoid a singularity...

    integral_2 += groove_field_z(x, y, lower_integration_limit + 0.01, lx, ly, lz);
    integral_2 += groove_field_z(x, y, upper_integration_limit - 0.01, lx, ly, lz);  
  
    for(i=1, z=lower_integration_limit + step_size; 
	i< npoints-1; 
	i++, z+=step_size) {
      temp = groove_field_z(x,y,z,lx,ly,lz);

      if(i%2 == 0) 
	integral_2 += 2.0 * temp;
      else 
	integral_2 += 4.0 * temp;
    }
    //integral_2 /= (3.0 * ((double) npoints-1));
    integral_2 *= step_size / 3.0;

    //return(integral_1 + integral_2);
    return((integral_1 + integral_2) / effective_material_height);
  }

  //
  // here we are not under the ridge...
  //

  lower_integration_limit = -1.0 * lz - base_material_height;
  upper_integration_limit = lower_integration_limit + effective_material_height;

  // compute the step size...
  step_size = (upper_integration_limit - lower_integration_limit) / ((double) npoints-1);
  
  integral_1  = groove_field_z(x,y,lower_integration_limit,        lx,ly,lz);
  integral_1 += groove_field_z(x,y,upper_integration_limit - 0.01, lx,ly,lz);
  
  unsigned i;
  for(i=1, z=lower_integration_limit + step_size; 
      i< npoints-1; 
      i++, z+=step_size) {
    temp = groove_field_z(x,y,z,lx,ly,lz);

    if(i%2 == 0) 
      integral_1 += 2.0 * temp;
    else 
      integral_1 += 4.0 * temp;
  }

  integral_1 /= (3.0 * ((double) npoints-1));

  return integral_1;

}
