/* dwm-groove-field.cc
   
   This file contains code to support the DWMGrooveField object. This object is designed to 
   compute and store the field produced by a rectangular grooving in the magnetic substrate.
   The groove is characterized by the following parameters:

   *put description here*


   =============================== CHANGE LOG ==================================

   1/10/94 added code in the object definition to individually set the groove width
   center, and depth.  This is part of a work around to solve the multiple groove
   initialization problem.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "dwm-v1.h"
#include "dwm-groove-field.h"
#include "dwm-debug.h"

extern struct MaterialParameters material_parameters;
extern struct InputMaterialParameters input_material_parameters;
extern struct RunParameters run_parameters;
extern struct InputRunParameters input_run_parameters;

void DWMGrooveField::initialize_with_defaults()
{
  groove_center_x = 0.0; groove_center_y = 0.0;
  groove_width_x = 0.0; groove_width_y = 0.0; groove_width_z = 0.0;

  border_width_x = 0.0; border_width_y = 0.0;

  lattice_constant_x = 100.0;  lattice_constant_y = 100.0;
  lattice_constant_product = 0.0;

  lattice_origin_x = 0.0; lattice_origin_y = 0.0;
  lattice_extent_x = 0.0; lattice_extent_y = 0.0;

  number_of_cells_x = 0; number_of_cells_y = 0;
  number_of_components = 0;
  
  description = 0; source = 0; data = 0;
}

void DWMGrooveField::reset()
{
  groove_center_x = 0.0; groove_center_y = 0.0;

  groove_width_x = 0.0; groove_width_y = 0.0; groove_width_z = 0.0;

  number_of_cells_x = 0; number_of_cells_y = 0; number_of_components = 0;

  lattice_constant_product = 0.0; lattice_constant_x = 0.0; 
  lattice_constant_y = 0.0;

  lattice_origin_x = 0.0; lattice_origin_y = 0.0;
  lattice_extent_x = 0.0; lattice_extent_y = 0.0;

  delete [] data;
  delete [] source;
  delete [] description;

}

void DWMGrooveField::allocate_storage_space()
{
  number_of_components = 1;
  unsigned nelements = number_of_cells_x * number_of_cells_y *  number_of_components;
  data = new double[nelements];
  for(int i = 0; i < nelements; i++) *(data+i) = 0.0;
}

void DWMGrooveField::remove_storage_space()
{
  delete [] data;
  delete [] source;
  delete [] description;
}

void DWMGrooveField::recompute_internals()
{
  lattice_constant_product = lattice_constant_x * lattice_constant_y;

  lattice_extent_x = lattice_origin_x + 
    lattice_constant_x * (number_of_cells_x - 1);

  lattice_extent_y = lattice_origin_y + 
    lattice_constant_y * (number_of_cells_y - 1);
}

double& DWMGrooveField::map_node(unsigned x, unsigned y, unsigned nc)
{
  if(x >= 0 && x < number_of_cells_x) { }
  else {
    printf("The value of x is %d\n",x);
    assert(x >= 0 && x < number_of_cells_x);
  }
  assert(y >= 0 && y < number_of_cells_y);
  return *(data + y * number_of_cells_x + x );
}

double DWMGrooveField::Interpolate(double x, double y, unsigned component)
{
  double a1, a2, a3, a4, result;

  // First check to see if we are outside the region that is
  // described by the mesh, and do something reasonable if this is
  // the case...

  if(x < lattice_origin_x || x >= lattice_extent_x) return 0.0;
  if(y < lattice_origin_y || y >= lattice_extent_y) return 0.0;

  // Get the mesh coordinates of the data point that is to the lower left 
  // of the specified coordinates.

//  unsigned xint = (int) floor((x - lattice_origin_x)/lattice_constant_x);
//  unsigned yint = (int) floor((y - lattice_origin_y)/lattice_constant_y);

  unsigned xint = (unsigned) floor((x - lattice_origin_x)/lattice_constant_x);
  unsigned yint = (unsigned) floor((y - lattice_origin_y)/lattice_constant_y);

  // x1 and y1 are the physical coordinates of the point described above.
  // They are used below in the interpolation algorithm.

  double x1 = xint *  lattice_constant_x + lattice_origin_x;
  double y1 = yint *  lattice_constant_y + lattice_origin_y;

  // Lookup the four adjacent mesh points...

  if(xint < 0 || xint >= number_of_cells_x) { 
    printf("Error: xint is out of range.\n");
    printf("xint = %u, yint = %u\n", xint, yint);
    printf("x    = %lf, y = %lf\n", x,y);
    printf("number_of_cells_x = %u, number_of_cells_y = %u\n",
	   number_of_cells_x, number_of_cells_y);
  }
  
  a1 = map_node(xint,   yint,   component);
  a2 = map_node(xint+1, yint,   component);
  a3 = map_node(xint,   yint+1, component);
  a4 = map_node(xint+1, yint+1, component);

  // ...and finally do the interpolation.
  
  result = a1 * ( 1.0 - (x-x1)/lattice_constant_x - (y-y1)/ lattice_constant_y +
		 (x-x1)*(y-y1)/(lattice_constant_x * lattice_constant_y) );

  result += a2 * ((x-x1)/ lattice_constant_x - 
		  (x-x1)*(y-y1)/(lattice_constant_x * lattice_constant_y));

  result += a3 * ((y-y1)/ lattice_constant_y - 
		  (x-x1)*(y-y1)/(lattice_constant_x * lattice_constant_y));

  result += a4 * (x-x1) * (y-y1) / ( lattice_constant_x * lattice_constant_y);
  return result;
}

//
// This procedure calculates the field produced by the groove or ridge as
// specified by the internal parameters that are stored in this object.
//

void DWMGrooveField::Reset()
{
  lattice_origin_x = groove_center_x - 0.5 * groove_width_x - border_width_x;
  lattice_origin_y = groove_center_y - 0.5 * groove_width_y - border_width_y;

  number_of_cells_x = (int)floor((2.0 * border_width_x + groove_width_x) / 
    lattice_constant_x) + 1;

  number_of_cells_y = (int)floor((2.0 * border_width_y + groove_width_y) / 
    lattice_constant_y) + 1;

  number_of_components = 1;

  recompute_internals();

  allocate_storage_space();  // allocate the space.

  unsigned i,j;
  double x_relative,y_relative, field;
  double x_absolute,y_absolute;
  double actual_material_thickness;

  // Here is where we call the calculation of the magnetic field.

  for(i=0; i < number_of_cells_x; i++) {
    for(j=0; j < number_of_cells_y; j++) {
    
      x_absolute = lattice_origin_x + i * lattice_constant_x;
      y_absolute = lattice_origin_y + j * lattice_constant_y;

      //x_relative = i * lattice_constant_x - 0.5 * groove_width_x - border_width_x;
      //y_relative = j * lattice_constant_y - 0.5 * groove_width_y - border_width_y;


      x_relative = (x_absolute - groove_center_x) * cos(groove_rotation_angle / 57.29578) -
	(y_absolute - groove_center_y) * sin(groove_rotation_angle / 57.29578);

      y_relative = (x_absolute - groove_center_x) * sin(groove_rotation_angle / 57.29578) +
	(y_absolute - groove_center_y) * cos(groove_rotation_angle / 57.29578);

      if(groove_width_z > 0.0) {
	actual_material_thickness = Unscaled_Material_Thickness(x_absolute, y_absolute);
      }

      else {
	if(x_relative < groove_width_x/2.0 &&
	   x_relative > -groove_width_x/2.0 &&
	   y_relative < groove_width_y/2.0 &&
	   y_relative > -groove_width_y/2.0) {
	  actual_material_thickness = input_material_parameters.material_thickness - groove_width_z;
	}
	else {
	  actual_material_thickness = input_material_parameters.material_thickness;
	}
      }
      if(groove_width_z > 0.0) {
	// this is a groove (reduced height) compared to the surrounding area.
	field = average_groove_field_z(x_relative,
				       y_relative,
				       input_material_parameters.material_thickness,
				       actual_material_thickness, 
				       0.5 * groove_width_x,
				       0.5 * groove_width_y, 
				       0.5 * groove_width_z);
	field *= -1.0;
	// also, note that we need a - sign here because the region of
	// material that is removes is opposite in direction to the bias field
	// (and opposite to the one given in this routine).
      }

      if(groove_width_z < 0.0) { 
	// this is a ridge (increased height) that it above the surrounding area.
	field = average_ridge_field_z(x_relative,
				      y_relative,
				      input_material_parameters.material_thickness,
				      actual_material_thickness, 
				      0.5 * groove_width_x,
				      0.5 * groove_width_y, 
				      -0.5 * groove_width_z);
      }

      field /=  12.566 ;  // take out the factor of four pi here so that I dont have to
                          // do it later when I lookup the field.
      map_node(i,j,1) = field;
    }
  }
}


void DWMGrooveField::Print_Diagnostics()
{
  printf("Groove Length along x: %f\n", groove_width_x);
  printf("Groove Length along y: %f\n", groove_width_y);
  printf("Groove Length along z: %f\n", groove_width_z);

  printf("Groove Center (x): %f\n", groove_center_x);
  printf("Groove Center (y): %f\n", groove_center_y);

  printf("No. of cells along x: %d\n", number_of_cells_x);
  printf("No. of cells along y: %d\n", number_of_cells_y);
  printf("No. of components:    %d\n", number_of_components);

  printf("X Lattice constant:       %f\n", lattice_constant_x);
  printf("Y Lattice constant:       %f\n", lattice_constant_y);
  printf("Lattice constant product: %f\n", lattice_constant_product);

  printf("X lattice origin:       %f\n", lattice_origin_x);
  printf("Y lattice origin:       %f\n", lattice_origin_y);

  printf("X lattice extent:       %f\n", lattice_extent_x);
  printf("Y lattice extent:       %f\n", lattice_extent_y);

  printf("Description: %s\n", description);
  printf("Source: %s\n", source);
}


void DWMGrooveField::Print_Data_Field()
{

  FILE *fp = fopen("field.data","w");
  double number;

/**********

  for(int j=0; j< number_of_cells_y; j++) {
      for(int i=0; i< number_of_cells_x; i++) {
	number = map_node(i,j,1);
	fprintf(fp,"%10.6f\n", number);
      }
      fprintf(fp,"\n");
    }
********************/

  for(double x=-5000.0; x< 5000.0; x+=100.0) {
      for(double y=-5000; y<5000.0; y+=100.0) {
	fprintf(fp,"%10.6f\n", Interpolate(x, y, 1));
      }
      fprintf(fp,"\n");
    }


  fclose(fp);
}

void DWMGrooveField::Write_Object_Description(FILE *fp) {
  fprintf(fp,"manhattangroove ");
  fprintf(fp,"limits %lf %lf %lf %lf ",  lattice_origin_x, lattice_origin_y, lattice_extent_x, lattice_extent_y);
  fprintf(fp,"data ");

  double x_original, y_original;
  double x_rotated, y_rotated;

  x_original = groove_center_x - 0.5 * groove_width_x;
  y_original = groove_center_y - 0.5 * groove_width_y;

  x_rotated = (x_original - groove_center_x) * cos(groove_rotation_angle / 57.29578) -
    (y_original - groove_center_y) * sin(groove_rotation_angle / 57.29578);
  y_rotated = (x_original - groove_center_x) * sin(groove_rotation_angle / 57.29578) +
    (y_original - groove_center_y) * cos(groove_rotation_angle / 57.29578);

  x_rotated += groove_center_x; y_rotated += groove_center_y;

  fprintf(fp,"%lf %lf ", x_rotated, y_rotated);
 

  x_original = groove_center_x + 0.5 * groove_width_x;
  y_original = groove_center_y - 0.5 * groove_width_y;

  x_rotated = (x_original - groove_center_x) * cos(groove_rotation_angle / 57.29578) -
    (y_original - groove_center_y) * sin(groove_rotation_angle / 57.29578);
  y_rotated = (x_original - groove_center_x) * sin(groove_rotation_angle / 57.29578) +
    (y_original - groove_center_y) * cos(groove_rotation_angle / 57.29578);

  x_rotated += groove_center_x; y_rotated += groove_center_y;

  fprintf(fp,"%lf %lf ", x_rotated, y_rotated);


  x_original = groove_center_x + 0.5 * groove_width_x;
  y_original = groove_center_y + 0.5 * groove_width_y;

  x_rotated = (x_original - groove_center_x) * cos(groove_rotation_angle / 57.29578) -
    (y_original - groove_center_y) * sin(groove_rotation_angle / 57.29578);
  y_rotated = (x_original - groove_center_x) * sin(groove_rotation_angle / 57.29578) +
    (y_original - groove_center_y) * cos(groove_rotation_angle / 57.29578);

  x_rotated += groove_center_x; y_rotated += groove_center_y;

  fprintf(fp,"%lf %lf ", x_rotated, y_rotated);


  x_original = groove_center_x - 0.5 * groove_width_x;
  y_original = groove_center_y + 0.5 * groove_width_y;

  x_rotated = (x_original - groove_center_x) * cos(groove_rotation_angle / 57.29578) -
    (y_original - groove_center_y) * sin(groove_rotation_angle / 57.29578);
  y_rotated = (x_original - groove_center_x) * sin(groove_rotation_angle / 57.29578) +
    (y_original - groove_center_y) * cos(groove_rotation_angle / 57.29578);

  x_rotated += groove_center_x; y_rotated += groove_center_y;

  fprintf(fp,"%lf %lf ", x_rotated, y_rotated);


  x_original = 	  groove_center_x - 0.5 * groove_width_x;
  y_original =   groove_center_y - 0.5 * groove_width_y;

  x_rotated = (x_original - groove_center_x) * cos(groove_rotation_angle / 57.29578) -
    (y_original - groove_center_y) * sin(groove_rotation_angle / 57.29578);
  y_rotated = (x_original - groove_center_x) * sin(groove_rotation_angle / 57.29578) +
    (y_original - groove_center_y) * cos(groove_rotation_angle / 57.29578);

  x_rotated += groove_center_x; y_rotated += groove_center_y;

  fprintf(fp,"%lf %lf ", x_rotated, y_rotated);

  fprintf(fp,"\n");

  /***************

  original code

  fprintf(fp,"%lf %lf ", 
	  groove_center_x - 0.5 * groove_width_x,
	  groove_center_y - 0.5 * groove_width_y);

  fprintf(fp,"%lf %lf ", 
	  groove_center_x + 0.5 * groove_width_x,
	  groove_center_y - 0.5 * groove_width_y);

  fprintf(fp,"%lf %lf ", 
	  groove_center_x + 0.5 * groove_width_x,
	  groove_center_y + 0.5 * groove_width_y);

  fprintf(fp,"%lf %lf ", 
	  groove_center_x - 0.5 * groove_width_x,
	  groove_center_y + 0.5 * groove_width_y);

  fprintf(fp,"%lf %lf ", 
	  groove_center_x - 0.5 * groove_width_x,
	  groove_center_y - 0.5 * groove_width_y);

  fprintf(fp,"\n");

*********************/

}



