/* dwm-main.cc
 * This is the main file for the non uniform distributed wall model code.
 * It contains the definition of all global variables, 
 * as well as the main() routine which does initialization, 
 * and all of the command processing routines.
 *
 *
 * Started 8/29/94 by gp.
 * Most of the code was initially stolen from test-4.lex
 * 
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <string.h>

#include "vector-1.h"
#include "wall-point-v3.h"
#include "domain-wall-v3.h"
#include "dwm-v1.h"
#include "dwm-dfield.h"
#include "dwm-flick.h"
#include "dwm-timing-obj.h"
#include "dwm-debug.h"
#include "dwm-keyword.h"
#include "dwm-groove-field.h"
#include "dwm-groove-region.h"

int yyparse(void);

/* this contains the basic, low level material parameters
   for distributed wall modelling.
   
   Note: only material_parameters and run_parameters are
   visible to the other parts of the simulation code.

*/

struct MaterialParameters material_parameters;
struct InputMaterialParameters input_material_parameters;

struct RunParameters run_parameters;
struct InputRunParameters input_run_parameters;


// this variable is initially set to false, as the domains have
// not yet been allocated.  Once this has been one time, we
// do not need to do it ever again, unless the user specifies that
// the simulation should be reset to its initial state at t=0.

DWMBoolean domain_walls_have_been_initialized = DWMFalse;

// Some globals that I dont know what to do with yet...

double input_gbias_field;
double gbias_field;
double gradient;
const double bubble_collapse_radius = 100.0;

Vector Xhat(3), Yhat(3), Zhat(3);

DWMFlickList *field_list;            // list of the external fields (not groove fields)
DWMFlickList *groove_field_list;     // list of the effective groove fields
DWMFlickList *groove_region_list;   // list of the grooved regions (to determine material thickness)
DWMFlickList *domain_wall_list;      // list of domain walls (pretty easy)
DWMFlickList *keyword_list;          // list of keywords (for command line processing)
DWMFlickList *global_object_list;    // list of everything (does not own elements)


extern FILE *yyin;

/*
 * Here, we begin the code to process sentences from the yacc grammar.
 * Each procedure has the name process_xxxx, where xxxx represents the
 * sentence name in the yacc grammar. Spme procedures only accept a few simple
 * typed arguments, while others accept a DWMFlickList of keyword options.
 * See the class DWMKeyword defined in dwm-keyword.h for the structure of these
 * keywords/value pairs.
 *
 */

void process_load_sentence(char *filename)
{
  FILE *old_yyin;
  FILE *new_yyin;

  if((new_yyin = fopen(filename,"r")) == NULL) {
    printf("\nLOAD ERROR: cannot find file named %s\n\n", filename);
    return;
  }

  // found the file, now lex and yacc it.
  // dont forget to save the current file pointer....

  old_yyin = yyin;

  yyin = new_yyin;  // now tell the lexer to look at the new text...

  while(!feof(yyin)) { 
    yyparse(); 
    keyword_list->Flush();
  }

  fclose(yyin);  // all done with the loaded file, so close it.

  yyin = old_yyin;  // and go back to whence we came!
}

void process_load_state_sentence(char *parameter)
{}

void process_load_field_sentence(char *field_name, char *source_name, DWMFlickList *options)
{}

void process_load_ascii_field_sentence(char *field_name, char *source_name, DWMFlickList *options)
{
  DWMDiscreteField *ptr;
  DWMBoolean error = DWMFalse;
  
  ptr = new DWMDiscreteField; // allocate a class to hold the field
  ptr->Rename(field_name);    // and give it a name for later use

  /* Now we process the load field keywords */

  DWMKeyword *foo;
  options->FReset();

  while( (foo = (DWMKeyword *)(options->Flick())) != DWMNoObject ) {
    if(!strcmp(foo->Name(), "xsize")) 
      { ptr->Set_Lattice_Size(0, (unsigned) foo->Dvalue()); continue; }
    if(!strcmp(foo->Name(), "ysize")) 
      { ptr->Set_Lattice_Size(1, (unsigned) foo->Dvalue()); continue; }

    if(!strcmp(foo->Name(), "xorigin")) 
      { ptr->Set_Lattice_Origin(0U, foo->Dvalue()); continue; }
    if(!strcmp(foo->Name(), "yorigin")) 
      { ptr->Set_Lattice_Origin(1U, foo->Dvalue()); continue; }

    if(!strcmp(foo->Name(), "xoffset")) {}
    if(!strcmp(foo->Name(), "yoffset")) {}

    if(!strcmp(foo->Name(), "dx")) 
      { ptr->Set_Lattice_Constant(0U, foo->Dvalue()); continue; }
    if(!strcmp(foo->Name(), "dy")) 
      { ptr->Set_Lattice_Constant(1U, foo->Dvalue()); continue; }
       
    printf("Error: Somehow I got a keyword that is unknown: %s\n", foo->Name());
    error = DWMTrue;
  }

  // if we have any errors, then abort the load...
  // otherwise, lets give it a go at this level...

  if(error == DWMTrue)  {
    printf("Aborting the load.\n");
    delete ptr;
    return;
  }

  ptr->Read_From_Ascii_File(source_name);

  if(run_parameters.debug) {
    printf("Discrete Field Diagnostics:\n");
    ptr->Print_Diagnostics();
  }

  field_list->Add(*ptr);
}


void process_create_bubble_sentence(char *name, DWMFlickList *options)
{
  // allocate a DomainWall class to hold the default position data and set the name..
  DomainWall *ptr; ptr = new DomainWall;

  if(name!=0) ptr->Rename(name);

  // next we process the option list information...

  DWMKeyword *key_ptr;  options->FReset();

  while( (key_ptr = (DWMKeyword *)(options->Flick())) != DWMNoObject ) {

    if(!strcmp(key_ptr->Name(), "xposition")) 
      { ptr->Set_Initial_X_Position(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "yposition")) 
      { ptr->Set_Initial_Y_Position(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "radius")) 
      { ptr->Set_Initial_Radius(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "npoints")) 
      { ptr->Set_Initial_Npoints((unsigned)key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "winding_number")) 
      { ptr->Set_Initial_Winding_Number(key_ptr->Ivalue()); continue; }

    if(!strcmp(key_ptr->Name(), "file")) 
      { 
	ptr->Set_Initial_State_Filename(key_ptr->Cvalue()); 
	continue; 
      }

    printf("Error: I have gotten the keyword: %s which I dont understand.\n",
	   key_ptr->Name());
  }

  domain_wall_list->Add(*ptr);

  global_object_list->Add(*ptr);

}

void process_create_tsequence_sentence(char *tsequence_name, 
				       char *field_name, 
				       DWMFlickList *options)
{
  DWMObject *ptr;
  DWMTimingObject *tobj_ptr;
  DWMDiscreteField *field_ptr;

  DWMBoolean error = DWMFalse;
  
  tobj_ptr = new DWMTimingObject;      // allocate a class to hold the field...
  tobj_ptr->Rename(tsequence_name);    // Give it a name for later use...

  // look up the named field...
  ptr = field_list->Contains_Named_Object(field_name);

  if(ptr == DWMNoObject) {
    printf("I cannot find the field name %s\n");
    delete tobj_ptr;
    return;
  }

  /* Now we have the specified field, so cast the pointer to a
     discrete field object so that we can add to it */

  field_ptr = (DWMDiscreteField *)ptr;

  /* Now we have the specified field and need to process the load 
     field keywords */

  DWMKeyword *key_ptr;
  options->FReset();

  while( (key_ptr = (DWMKeyword *)(options->Flick())) != DWMNoObject ) {
    if(!strcmp(key_ptr->Name(), "rise_time")) 
      { tobj_ptr->Set_Rise_Time(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "fall_time")) 
      { tobj_ptr->Set_Fall_Time(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "pulse_width")) 
      { tobj_ptr->Set_Pulse_Width(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "pulse_cycle")) 
      { tobj_ptr->Set_Cycle_Time(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "pulse_offset")) 
      { tobj_ptr->Set_Pulse_Offset(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "bias_current")) 
      { tobj_ptr->Set_Bias_Current(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "pulse_current")) 
      { tobj_ptr->Set_Pulse_Current(key_ptr->Dvalue()); continue; }

    printf("I dont know %s\n", key_ptr->Name());
  }

  if(error == DWMTrue) {
    printf("Aborting tsequence generation.\n");
    delete tobj_ptr;
    return;
  }
  // ok. now commit.

  field_ptr->Add_TimingObject(*tobj_ptr);

  // and save it on a global list for future access

  global_object_list->Add(*tobj_ptr);

  if(run_parameters.debug) {
    printf("Timing Sequence Created:\n");
    tobj_ptr->Print_Diagnostics();
  }
}

void process_create_groove_sentence(char *name, 
				    double xc, double yc, 
				    double xw, double yw, 
				    double zw, DWMFlickList *options)
{
  DWMGrooveField *gfield_ptr;       gfield_ptr = new DWMGrooveField;
  DWMManhattanRegion *gregion_ptr;  gregion_ptr = new DWMManhattanRegion;

  if(name!=0) {
    gfield_ptr->Rename(name);
    gregion_ptr->Rename(name);
  }

  // check the values that have been passed to this routine so as to catch
  // any trivial errors

  if( xw <= 0.0 || yw <= 0.0) { 
    printf("Error: You cannot specify a groove with a negative or zero width parameter.\n");
    delete gfield_ptr; 
    delete gregion_ptr;
    return; 
  }

  if( zw == 0.0 ) { 
    printf("Error: You cannot specify a groove with a zero depth/height\n");
    delete gfield_ptr; 
    delete gregion_ptr;
    return; 
  }

  // 1/10/94 -- replacing initialize call with the following set commands
  // gfield_ptr->Initialize( xc, yc, xw, yw, zw,  1.0);

  gfield_ptr->Set_X_Center(xc);
  gfield_ptr->Set_Y_Center(yc);
  gfield_ptr->Set_X_Width(xw);
  gfield_ptr->Set_Y_Width(yw);
  gfield_ptr->Set_Z_Width(zw);

  // choose the xborder and yborder values automatically here.  If the
  // used has specified them then they will be picked up below...

  double border_width_x = 2.0 * xw;
  double border_width_y = 2.0 * yw;

  if(border_width_y < border_width_x) border_width_x = border_width_y;
  else border_width_y = border_width_x;

  gfield_ptr->Set_X_Border(border_width_x);
  gfield_ptr->Set_Y_Border(border_width_y);
    
  // process the option list information...
  DWMKeyword *key_ptr;  options->FReset();

  while((key_ptr = (DWMKeyword *)(options->Flick())) != DWMNoObject ) {
    if(!strcmp(key_ptr->Name(), "xborder")) 
      { gfield_ptr->Set_X_Border(key_ptr->Dvalue()); continue; }
 
    if(!strcmp(key_ptr->Name(), "yborder")) 
      { gfield_ptr->Set_Y_Border(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "dx")) 
      { gfield_ptr->Set_X_Lattice_Constant(key_ptr->Dvalue()); continue; }

    if(!strcmp(key_ptr->Name(), "dy")) 
      { gfield_ptr->Set_Y_Lattice_Constant(key_ptr->Dvalue()); continue; }
 
   if(!strcmp(key_ptr->Name(), "rotation")) 
      { gfield_ptr->Set_Rotation_Angle(key_ptr->Dvalue()); continue; }
 
    printf("Error creating groove: I dont know the keyword %s\n", 
	   key_ptr->Name());
  }

  // dont forget to initialize the region object.
  gregion_ptr->Initialize( xc, yc, xw, yw, zw);

  // add the objects to their respective lists.
  groove_field_list->Add(*gfield_ptr);
  groove_region_list->Add(*gregion_ptr);

  // finally print diagnotics info if debug is set.
  if(run_parameters.debug) {
    gfield_ptr->Print_Diagnostics();
    gregion_ptr->Print_Diagnostics();
  }
}

void process_simple_set_sentence(char *parameter_name, DWMfloat value)
{
  if( strcmp(parameter_name, "ms") == 0)
    { input_material_parameters.four_pi_ms = value; return; }

  if( strcmp(parameter_name, "clength") == 0)
    { input_material_parameters.characteristic_length    = value; return; }

  if( strcmp(parameter_name, "alpha") == 0)
    { input_material_parameters.alpha = value; return; }

  if( strcmp(parameter_name, "q") == 0)
    { input_material_parameters.quality_factor        = value; return; }

  if( strcmp(parameter_name, "mobility") == 0)
    { input_material_parameters.wall_mobility         = value; return; }

  if( strcmp(parameter_name, "maximum_velocity") == 0) 
    { input_material_parameters.maximum_wall_velocity = value; return; }

  if( strcmp(parameter_name, "thickness") == 0) 
    { input_material_parameters.material_thickness  = value; return; }

  if( strcmp(parameter_name, "time_step") == 0) 
    { input_run_parameters.time_step = value; return; }

  if( strcmp(parameter_name, "time_save") == 0) 
    { input_run_parameters.state_save_interval = value; return; }

 if( strcmp(parameter_name, "bias") == 0) 
   { input_gbias_field = value; return; }

 if( strcmp(parameter_name, "coercivity") == 0)
   { input_material_parameters.coercivity = value; return; }

 if( strcmp(parameter_name, "debug") == 0) 
   { run_parameters.debug = (int)value; return; }

  printf("I dont know how to set %s to value %f\n", parameter_name, value);

}

void process_simple_set_sentence(char *parameter_name, char *value)
{
  if(run_parameters.debug == 2)
    printf("process_simple_set_sentence: arg1 = %s arg2 = %s\n", parameter_name, value);

 if( strcmp(parameter_name, "output_name") == 0) 
   { strcpy(run_parameters.base_name, value); return; }

 if( strcmp(parameter_name, "debug") == 0) {
   if( strcmp(value,"on") == 0) { run_parameters.debug = 1; return; }
   if( strcmp(value,"off") == 0) { run_parameters.debug = 0; return; }
   printf("Error: I dont know SET debug %s\n", value);
   return;
 }
  printf("I dont know how to set %s to value %s\n", parameter_name, value);
}

void process_delete_sentence(char *name_a, char *name_b = 0)
{
  // here we are deleting a whole object...
  if( name_b == 0) { 

    // Gotta look for the object. First check the walls.

    DomainWall *dw_ptr;
    if((dw_ptr = (DomainWall *)domain_wall_list->Contains_Named_Object(name_a)) != DWMNoObject) {
      printf("Deleting domain wall named %s\n", dw_ptr->Name());
      //delete dw_ptr; 
      domain_wall_list->Detach(*dw_ptr);
      return;
    }

    // check the loaded external fields...

    DWMField *field_ptr;
    if((field_ptr = (DWMField *)field_list->Contains_Named_Object(name_a)) != DWMNoObject) {
      printf("Deleting external field named %s\n", field_ptr->Name());
      //delete field_ptr; 
      field_list->Detach(*field_ptr);
      return;
    }

    // ...check the grooves....
    DWMGrooveField *gfield_ptr;
    DWMGrooveRegion *gregion_ptr;

    if((gfield_ptr = (DWMGrooveField *)groove_field_list->Contains_Named_Object(name_a)) != DWMNoObject) {

      // first check to see that we can find the appropriate region.
      if((gregion_ptr = (DWMGrooveRegion *)groove_region_list->Contains_Named_Object(name_a)) == DWMNoObject) {
	printf("Error: I found the field named %s, but I cannot find a region associated with it.\n");
	return;
      }
      
      // here we have found both structures...
      printf("Deleting groove field named %s\n", gfield_ptr->Name());

      groove_field_list->Detach(*gfield_ptr);
      groove_region_list->Detach(*gregion_ptr);
      return;
    }

//    DWMTimingObject *tobj;
//    field_list->FReset();
//    while((field_ptr = (DWMDiscreteField *)field_list->Flick()) != DWMNoObject &&
//	  (tobj = (DWMTimingObject *)field_ptr->Contains_Named_Object(name_a)) != DWMNoObject) {
      // found it -- now remove it.
  //    printf("Deleting timing sequence named %s\n", tobj->Name());
    //  printf("Fix this code!!!\n");
      //field_object->Detach(*field_ptr);
      //return;
    //}
  }

  // Here we are deleting items that require two names to locate them....
  // do we have any of these ?

}

void process_run_sentence(DWMfloat run_length)
{
  DWMfloat my_run_length;
  DomainWall *ptr;

  // first thing that we do is to recompute the internal parameters...
  calculate_internal_parameters();

  // If necessary, do a Reset() on all of the domain walls that we are simulating....
  if(domain_walls_have_been_initialized == DWMFalse) {
    DomainWall *ptr;
    domain_wall_list->FReset();
    while( (ptr = (DomainWall*)domain_wall_list->Flick()) != DWMNoObject ) ptr->Reset();

    // put the groove initialization code here for now.
    // loop over all of the groove field objects and Reset() them...
    DWMObject *optr;
    DWMGrooveField *gf_ptr;

    groove_field_list->FReset();
    while((optr = groove_field_list->Flick()) != DWMNoObject ) {
      gf_ptr = (DWMGrooveField *)optr;
      printf("Resetting groove: %s\n", gf_ptr->Name());
      gf_ptr->Reset();
    }
    printf("Groove reset done.\n");
    domain_walls_have_been_initialized = DWMTrue;
  }

  // ...and if the streams have not been opened, then initialize them.

  if(run_parameters.state_fp == 0) { reset_file_pointers();
				     setup_output_file(run_parameters.state_fp);
				   }

  // See if a time argument is specified. If not, then use the previous value...
  
  if(run_length == -1.0) 
    my_run_length = input_run_parameters.total_time_requested;
  else {
    my_run_length = run_length;
    input_run_parameters.total_time_requested = run_length;
  }

  // next, we calculate how many steps to take (an integer), and also how often 
  // to write out the state info.

  int nsteps = (int)(my_run_length / input_run_parameters.time_step);
  int summary_steps = (int)(run_parameters.state_save_interval /
			    run_parameters.time_step);
  int dw_n_elements;
  int dw_count;
  int i;

  dw_n_elements = domain_wall_list->Number_Of_Elements();
  if( dw_n_elements == 0 ) {
    printf("There are no domain walls to simulate.\n");
    return;
  }

  // loop over time index...
  for(i=0; i<nsteps; i++) {

    // check the number of domain walls, as it may have changed...
    dw_n_elements = domain_wall_list->Number_Of_Elements();
    if( dw_n_elements == 0 ) {
      printf("No domain walls remaining in the list. Time = %f\n", 
	     run_parameters.time * 1.0);
      return;
    }

    domain_wall_list->FReset();

    // loop over domain walls.....
    for(dw_count = 0; dw_count < dw_n_elements; dw_count++) {
      ptr = (DomainWall *)domain_wall_list->Flick();
      ptr->Examine_And_Correct_Shape();
      ptr->Update_Fields();
      ptr->Update_Internals();
      ptr->Calculate_Demag_Field();
      ptr->Calculate_Forces();
      ptr->Calculate_Delta_R();
      ptr->Calculate_Delta_S();
      ptr->Calculate_Delta_Phia();
      ptr->Update_Position();
    }

    // update the global time and step count.
    run_parameters.time += run_parameters.time_step;
    run_parameters.step++;

    // code to write bubble/domain wall states out to a file...
    if( i%summary_steps == 0) {
      domain_wall_list->FReset();

      fprintf(run_parameters.state_fp,"state %d %f ", i, run_parameters.time * 
	 material_parameters.time_scale_factor); 

      for(dw_count = 0; dw_count < dw_n_elements; dw_count++) {
	ptr = (DomainWall *)domain_wall_list->Flick();
	ptr->Print_State_And_Phia(run_parameters.state_fp);
      }
      
      fprintf(run_parameters.state_fp, "\n");
    }

    // Check to see if all of the walls are at equilibrium.
    Vector max_disp;
    domain_wall_list->FReset();
    int domains_at_equilibrium = 0;

    for(dw_count = 0; dw_count < dw_n_elements; dw_count++) {
      ptr = (DomainWall *)domain_wall_list->Flick();
      max_disp = ptr->Maximum_Wall_Point_Displacement();

      if(max_disp.Get_Component(0) < 1.0e-5 &&
	 max_disp.Get_Component(1) < 1.0e-5 &&
	 max_disp.Get_Component(2) < 1.0e-5) 
	domains_at_equilibrium++;
    }

//    if(domains_at_equilibrium == dw_n_elements) {
//      printf("The domains are at equilibrium\n");
//      break;
//    }

    // Check to see if bubbles need to be removed....
    domain_wall_list->FReset();
    for(dw_count = 0; dw_count < dw_n_elements; dw_count++) {
      ptr = (DomainWall *)domain_wall_list->Flick();
      if( ptr->Domain_Radius() * material_parameters.distance_scale_factor 
	 < bubble_collapse_radius) {
	printf("Domain %s has collapsed.\n", ptr->Name());
	//domain_wall_list->Detach( *ptr );
	goto loop_exit;
      }
    }
  }

 loop_exit:

  // Flush all of the buffers and let the user know that the run is done...
  fflush(run_parameters.state_fp);
  fflush(run_parameters.debug_fp);
  fflush(run_parameters.summary_fp);
  printf("done.\n");
}


void reset_file_pointers()
{
  // First close any files that are open...

  if(run_parameters.debug_fp) fclose(run_parameters.debug_fp);
  if(run_parameters.summary_fp) fclose(run_parameters.summary_fp);
  if(run_parameters.state_fp) fclose(run_parameters.state_fp);

  // ...create new file names...

  char debug_name[256];
  char state_name[256];
  char summary_name[256];

  sprintf(debug_name,"%s.debug", run_parameters.base_name);
  sprintf(state_name,"%s.dwm", run_parameters.base_name);
  sprintf(summary_name,"%s.sum", run_parameters.base_name);

  // ....open them....
  run_parameters.debug_fp   = fopen(debug_name,  "w");
  run_parameters.summary_fp = fopen(summary_name,"w");
  run_parameters.state_fp   = fopen(state_name,  "w");
}




void process_reset_sentence()
{
  // Easy. First we recalculate the internal parameters...

  calculate_internal_parameters();

  // Pass a reset statement on to all domain walls...
  
  reset_domain_walls();
  
  // Reset the file pointers:

  reset_file_pointers();

  // setup the output file with the grooving structure...

  setup_output_file(run_parameters.state_fp);

  // Finally reset the global time and global step count to zero.

  run_parameters.time = 0.0;
  run_parameters.step = 0;
}

void process_print_sentence(char *print_arg_1 , char *print_arg_2)
{
  // first we deal with calls that look like 'print' and 'print quantity'
  if( print_arg_1 == 0 && print_arg_2 == 0) { print_parameters(); return; }
  if( strncmp(print_arg_1,"para",4) == 0)   { print_parameters(); return; }     // print parameters
  if( strncmp(print_arg_1,"wall",4) == 0) { print_domain_walls(); return; }     // print walls
  if( strncmp(print_arg_1,"fiel",4) == 0) { print_external_fields(); return; }  // print fields
  if( strncmp(print_arg_1,"obje",4) == 0) {  return; }  // print objects -- NOT YET IMPLEMENTED

  // next we look for calls like 'print object quantity'
  // first look at domain walls.

  DomainWall *dw_ptr;
  if((dw_ptr = (DomainWall *)domain_wall_list->Contains_Named_Object(print_arg_1)) != DWMNoObject) {

    printf(" Diagnostic information on domain named %s\n", dw_ptr->Name());
    
    if(print_arg_2 == 0) dw_ptr->Print_Diagnostics();
    else if ( strcmp(print_arg_2, "diagnostics") == 0) dw_ptr->Print_Diagnostics(); 
    else if ( strcmp(print_arg_2, "diagnostics1") == 0) dw_ptr->Print_Diagnostics_1(); 
    else if ( strcmp(print_arg_2, "diagnostics2") == 0) dw_ptr->Print_Diagnostics_2(); 
    else if ( strcmp(print_arg_2, "diagnostics3") == 0) dw_ptr->Print_Diagnostics_3(); 
    else if ( strcmp(print_arg_2, "fields") == 0) dw_ptr->Print_Fields(); 
    return;
  }


}

void process_comment_sentence(DWMFlickList *options)
{
  // what should we do with the comments ? Lets just have look at them here.

  DWMKeyword *key_ptr;  options->FReset();

  printf("COMMENT: ");

  while( (key_ptr = (DWMKeyword *)(options->Flick())) != DWMNoObject ) {
    printf("%s ", key_ptr->Cvalue());
  }
  printf("\n");
}

void process_quit_sentence()
{
  fclose(run_parameters.state_fp);
  fclose(run_parameters.summary_fp);
  fclose(run_parameters.debug_fp);
  exit(0);
}

main(int argc, char **argv)
{
  /********************* First we do some setup ***************************/
  /* Here are the basis vectors for the coordinate system */

  Vector temp(3);
  Xhat = temp; Yhat = temp; Zhat = temp;
  Xhat.Set_Component(0,1.0);
  Yhat.Set_Component(1,1.0);
  Zhat.Set_Component(2,1.0);


  /* ....allocate the list structures.....*/

  field_list          = new DWMFlickList;
  groove_field_list   = new DWMFlickList;
  groove_region_list  = new DWMFlickList;
  domain_wall_list    = new DWMFlickList;
  keyword_list        = new DWMFlickList;
  global_object_list  = new DWMFlickList;

  field_list->Owns_List_Elements()          = DWMTrue;
  groove_field_list->Owns_List_Elements()   = DWMTrue; 
  groove_region_list->Owns_List_Elements()  = DWMTrue; 
  domain_wall_list->Owns_List_Elements()    = DWMTrue;
  keyword_list->Owns_List_Elements()        = DWMTrue;
  global_object_list->Owns_List_Elements()  = DWMFalse;

  /* ... and setup some reasonable material parameters.... */

  input_material_parameters.four_pi_ms = 179.0;
  input_material_parameters.characteristic_length = 556.0;
  input_material_parameters.quality_factor = 7.84;
  input_material_parameters.alpha = 0.1;
  input_material_parameters.wall_mobility = 5.80;
  input_material_parameters.maximum_wall_velocity = 58.0;
  input_material_parameters.material_thickness = 5250.0;
  input_material_parameters.coercivity = 0.0;

  material_parameters.four_pi_ms = 0.0;
  material_parameters.quality_factor = 0.0;
  material_parameters.alpha = 0.0;
  material_parameters.wall_mobility = 0.0;
  material_parameters.maximum_wall_velocity = 0.0;
  material_parameters.material_thickness = 0.0;
  material_parameters.coercivity = 0.0;
  material_parameters.maximum_wall_force = 0.0;
  material_parameters.distance_scale_factor = 0.0;
  material_parameters.time_scale_factor = 0.0;
  material_parameters.magnetic_field_scale_factor = 0.0;

  input_run_parameters.time_step = 1.0;
  input_run_parameters.total_time_requested = 100.0;
  input_run_parameters.state_save_interval = 10.0;
  strcpy(run_parameters.base_name, "default");

  run_parameters.time_step = 0.0;
  run_parameters.total_time_requested = 0.0;
  run_parameters.state_save_interval = 0.0;
  run_parameters.time = 0.0;
  run_parameters.step = 0;
  run_parameters.debug_fp = 0;
  run_parameters.state_fp = 0;
  run_parameters.summary_fp = 0;

  input_gbias_field = 85.0;

  while(!feof(yyin)) { 
    yyparse(); 
    keyword_list->Flush();
  }
}

//
//
// The following code contains some general purpose utility procedures
// and are usually called from the command processing code.  One exception
// is with Material_Thickness(...) which is called from the demag calculation
// routine to evaluate the demag field related integrals.
//
//
//

void calculate_internal_parameters() 
{
  // First we re-calculate material parameters so that we can reset them during a run.

  material_parameters.distance_scale_factor = 
    input_material_parameters.characteristic_length / 
      (2.0 * input_material_parameters.quality_factor);

  material_parameters.magnetic_field_scale_factor = 
    input_material_parameters.four_pi_ms;

  double mygamma = 2.0 * input_material_parameters.wall_mobility *
    input_material_parameters.alpha * input_material_parameters.quality_factor / 
      input_material_parameters.characteristic_length;

  material_parameters.time_scale_factor = 
    1.0/(input_material_parameters.four_pi_ms * mygamma);

  material_parameters.maximum_wall_force = 
    input_material_parameters.maximum_wall_velocity /
      (input_material_parameters.four_pi_ms * input_material_parameters.wall_mobility);

  material_parameters.four_pi_ms = input_material_parameters.four_pi_ms;
  material_parameters.quality_factor = input_material_parameters.quality_factor;
  material_parameters.alpha = input_material_parameters.alpha;

  material_parameters.material_thickness = 
    input_material_parameters.material_thickness / material_parameters.distance_scale_factor;

  material_parameters.coercivity = 
    input_material_parameters.coercivity / 
      material_parameters.magnetic_field_scale_factor;

  run_parameters.time_step = 
    input_run_parameters.time_step / material_parameters.time_scale_factor;

  run_parameters.total_time_requested = 
    input_run_parameters.total_time_requested / material_parameters.time_scale_factor;
  
  run_parameters.state_save_interval = 
    input_run_parameters.state_save_interval / material_parameters.time_scale_factor;

  gbias_field = input_gbias_field / material_parameters.magnetic_field_scale_factor;

//  strcpy(run_parameters.base_name, input_run_parameters.base_name);

}


void print_parameters()
{
  printf("\t4piMs = \t\t%lf\n",            input_material_parameters.four_pi_ms);
  printf("\tCharacteristic Length = %lf\n",input_material_parameters.characteristic_length);
  printf("\tQuality Factor = \t%lf\n",     input_material_parameters.quality_factor);
  printf("\tAlpha = \t\t%lf\n",            input_material_parameters.alpha);
  printf("\tWall Mobility = \t%lf\n",      input_material_parameters.wall_mobility);
  printf("\tMax Wall Velocity = \t%lf\n",  input_material_parameters.maximum_wall_velocity);
  printf("\tMaterial Thickness = \t%lf\n", input_material_parameters.material_thickness);
  printf("\tCoercivity = \t\t%lf\n",         input_material_parameters.coercivity);
  printf("\tTime Step = \t\t%lf\n",        input_run_parameters.time_step);
  printf("\tTotal Time Requested  = %lf\n",input_run_parameters.total_time_requested);
  printf("\tState Save Interval = \t%lf\n",input_run_parameters.state_save_interval);
  printf("\tDebug  = \t\t%d\n",           run_parameters.debug);
  printf("\tBias Field  = \t\t%lf\n",      input_gbias_field);
  printf("\tTime  = \t\t%lf\n",            run_parameters.time * 
	 material_parameters.time_scale_factor);
  printf("\tGradient  = \t\t%lf\n",gradient);

//  printf("\tDistance Scale Factor = \t%lf\n",material_parameters.distance_scale_factor);
//  printf("\tTime Scale Factor = \t%lf\n",material_parameters.time_scale_factor);
//  printf("\tMagnetic Field Scale Factor = \t%lf\n", 
//	 material_parameters.magnetic_field_scale_factor);

}

void print_domain_walls()
{  
  int count, index;
  DomainWall *ptr;

  count = domain_wall_list->Number_Of_Elements();
  printf("------------------------------ DOMAIN WALL LIST ------------------------------\n");
  printf("There are currently %d domain walls walls.\n\n", count);

  domain_wall_list->FReset();
  index = 0;
  while( (ptr = (DomainWall*)domain_wall_list->Flick()) != DWMNoObject ) {
    printf("Index = %2d  Name = %10s  Position = (%10.6f, %10.6f)  Radius = %f\n",
	   index, ptr->Name(), 
	   ptr->Domain_Position().Get_Component(0) * material_parameters.distance_scale_factor,
	   ptr->Domain_Position().Get_Component(1) * material_parameters.distance_scale_factor,
	   ptr->Domain_Radius() * material_parameters.distance_scale_factor);
    index++;
  }
  printf("\n\n");
}

void print_external_fields()
{
  int count = field_list->Number_Of_Elements();
  DWMField *ptr;

  printf("EXTERNAL FIELD LIST: %d elements.\n", count);

  field_list->FReset();
  for(int i=0; i<count; i++) {
    ptr = (DWMField *)(field_list->Flick());
    printf("index = %2d  Name = %10s  \n", i, ptr->Name());
  }
  printf("\n\n");
}

// returns the absolute material thickness in nm, not internal units...

DWMfloat Unscaled_Material_Thickness(DWMfloat x, DWMfloat y)
{
  DWMGrooveRegion *ptr;
  int count = groove_region_list->Number_Of_Elements();
  DWMfloat groove_depth = 0.0;

  groove_region_list->FReset();  // Reset the flick list

  for(int i=0; i<count; i++) {
    ptr = (DWMGrooveRegion *)(groove_region_list->Flick());

    if(ptr->In_Region(x, y) == DWMTrue) {
      groove_depth = ptr->Groove_Depth(x,y);
      break;
    }
  }
  return(input_material_parameters.material_thickness - groove_depth);
}

DWMfloat Material_Thickness(DWMfloat x, DWMfloat y)
{
  DWMGrooveRegion *ptr;
  int count = groove_region_list->Number_Of_Elements();
  double scaled_groove_depth = 0.0;

  groove_region_list->FReset();  // Reset the flick list

  for(int i=0; i<count; i++) {
    ptr = (DWMGrooveRegion *)(groove_region_list->Flick());

    // if this point is inside a grooved region, then
    // get the depth of the groove, scale it by the appropriate distance scale factor
    // (recall that it will be stored in nanometers)
    // and break out of the loop -- i.e. use the data for the first 
    // grooved region that we find (that encompasses this point).

    if(ptr->In_Region(x, y) == DWMTrue) {
      scaled_groove_depth = ptr->Groove_Depth(x,y) / material_parameters.distance_scale_factor;
      break;
    }
  }

  return(material_parameters.material_thickness - scaled_groove_depth);
}

DWMfloat Material_Thickness(Vector &location)
{
  // OK -- so I am lazy.  Here it is:
  DWMfloat x = location.Get_Component(0);
  DWMfloat y = location.Get_Component(1);
  return Material_Thickness(x,y);
}

void reset_domain_walls()
{
    DomainWall *ptr;
    domain_wall_list->FReset();
    while( (ptr = (DomainWall*)domain_wall_list->Flick()) != DWMNoObject ) ptr->Reset();
}

void set_all_timing_object_currents(DWMfloat value)
{

  DWMObject *ptr;
  DWMTimingObject *to_ptr;

  global_object_list->FReset();

  // loop over all of the objects. Pull out the timing objects and set the currents appropriately.

  while((ptr = global_object_list->Flick()) != DWMNoObject ) {
    if(ptr->Class_Type() == DWMTimingObjectClass) {
      to_ptr = (DWMTimingObject *)ptr;
      //printf("setting current in timing object named %s\n", to_ptr->Name());
      if(to_ptr->Pulse_Current() > 0.0) to_ptr->Set_Pulse_Current( value );
      if(to_ptr->Pulse_Current() <  0.0) to_ptr->Set_Pulse_Current( - value );
    }
  }
}


void setup_output_file(FILE *fp) {
  // this procedure writes out all of the 'initialization' objects and material parameters
  // to a file.

  // first the parameters...
  fprintf(fp, "parameter 4piMs %lf\n",         input_material_parameters.four_pi_ms);
  fprintf(fp, "parameter Clength %lf\n",       input_material_parameters.characteristic_length);
  fprintf(fp, "parameter Q %lf\n",             input_material_parameters.quality_factor);
  fprintf(fp, "parameter Alpha %lf\n",         input_material_parameters.alpha);
  fprintf(fp, "parameter WallMobility %lf\n",  input_material_parameters.wall_mobility);
  fprintf(fp, "parameter MaximumWallVelocity %lf\n",
	  input_material_parameters.maximum_wall_velocity);
  fprintf(fp, "parameter MaterialThickness %lf\n",input_material_parameters.material_thickness);
  fprintf(fp, "parameter Coercivity %lf\n",    input_material_parameters.coercivity);
  fprintf(fp, "parameter TStep %lf\n",         input_run_parameters.time_step);
  fprintf(fp, "parameter TSave %lf\n",         input_run_parameters.state_save_interval);
  fprintf(fp, "parameter Debug %d\n",          run_parameters.debug);
  fprintf(fp, "parameter BiasField  %lf\n",    input_gbias_field);

  // now the groove fields
  DWMObject *ptr;
  DWMGrooveField *gf_ptr;

  groove_field_list->FReset();

  // loop over all of the groove field objects and write out their descriptions...

  while((ptr = groove_field_list->Flick()) != DWMNoObject ) {
    gf_ptr = (DWMGrooveField *)ptr;
    gf_ptr->Write_Object_Description(fp);
  }


  // now the field patterns...
  // loop over all of the field objects and write out their descriptions...
  DWMDiscreteField *df_ptr;

  field_list->FReset();
  while((ptr = field_list->Flick()) != DWMNoObject ) {
    printf("calling write object.\n");
    df_ptr = (DWMDiscreteField *)ptr;
    df_ptr->Write_Object_Description(fp);
  }
}


void process_write_object_sentence(char *object_name, char *file_name)
{
  DWMObject *ptr;
  
  if((ptr = global_object_list->Contains_Named_Object(object_name)) == DWMNoObject) {
    printf("Error: cannot find object named %s\n", object_name);
    return;
  }

  // ok. found the object. open a file and write it out....

  FILE *fp;
  if((fp = fopen(file_name, "w")) == NULL) {
    printf("Error: I cannot open the file named %s.\n", file_name);
    return;
  }

  if( ptr->Class_Type() == DWMDomainWallClass ) ((DomainWall *)ptr)->Write(fp);

  else {
    printf("Error: I do not know how to ideentify the type of this object.\n");
  }

  fclose(fp);
}

void process_write_field_sentence(char *fname, 
				  DWMFloat xlow,  DWMFloat xhigh,  
				  DWMFloat ylow,  DWMFloat yhigh,  
				  DWMFloat dx,    DWMFloat dy)
{

  FILE *fp;

  fp = fopen(fname, "w");

  double x, y;
  double field;


  for(y=ylow;  y<=yhigh;   y+= dy) {

    for(x=xlow; x<= xhigh; x+= dx) {
      field = 0.0;

      // x1,y1 is the position in nanometers
  
      double real_time = run_parameters.time * material_parameters.time_scale_factor;

      int count = field_list->Number_Of_Elements();
      field_list->FReset();
      int i;

      for(i=0; i<count; i++) {
	field += ((DWMField *)(field_list->Flick()))->Evaluate_Field(x,y,real_time) / 
	  material_parameters.four_pi_ms;
      }

      count = groove_field_list->Number_Of_Elements();
      groove_field_list->FReset();

      // note that the magnetic fields stored in groove_field_list do not have to be 
      // reduced by 4 pi Ms

      for(i=0; i<count; i++) {
	field += ((DWMField *)(groove_field_list->Flick()))->Evaluate_Field(x,y,real_time);
      }

      fprintf(fp,"%lf ", field);
    }
    fprintf(fp,"\n");
  }
  fclose(fp);
}
