#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <string.h>

#include "vector-1.h"
#include "wall-point-v3.h"
#include "domain-wall-v3.h"
#include "dwm-v1.h"
#include "dwm-flick.h"
#include "dwm-debug.h"

extern struct MaterialParameters material_parameters;
extern struct RunParameters run_parameters;
extern struct InputRunParameters input_run_parameters;

extern DWMFlickList *field_list;          
extern DWMFlickList *groove_field_list;
extern DWMFlickList *groove_region_list;
extern DWMFlickList *domain_wall_list;
extern DWMFlickList *keyword_list;
extern DWMFlickList *global_object_list;

extern double gbias_field;
extern double input_gbias_field;
//extern const double bubble_collapse_radius;

enum rte1_status { error, equilibrium, done, collapse, stripe_out };

int run_utility(double);
rte1_status run_to_equilibrium_1(double foo  = 0.0);

void ml_optimize(double run_time, double initial_current, double ytest, 
		 double x_low_test, double x_high_test)
{
  int run_count; 
  double current;
  DomainWall *ptr;
  Vector bubble_position; double bubble_position_x, bubble_position_y;

  double lower_current = -1.0;
  double upper_current = -1.0;
  double current_increment = 5.0;

  double tolerance = 0.1;
  int max_run_count = 20;
  int status = 0;

  char local_base_name[64];
  strcpy(local_base_name, run_parameters.base_name);  // save the initially specified name

  // search up until we bracket the point of interest.

  printf("Searching for device operation...\n");

  current = initial_current;

  for(run_count = 0; run_count < max_run_count; run_count++) {

    sprintf(run_parameters.base_name, "%s-%2.2f", local_base_name, current);
    printf("Current Increment = %f\n", current_increment);
    reset_file_pointers();
    setup_output_file(run_parameters.state_fp);
    reset_domain_walls();
    run_parameters.time = 0.0;
    run_parameters.step = 0;
    set_all_timing_object_currents(current);  // set the currents

    // do the run...

    printf("Run: name = %s current = %f count = %d\n",
	   run_parameters.base_name, current, run_count);

    status = run_utility(run_time);

    // next check the bubble position....
    domain_wall_list->FReset(); 
    ptr = (DomainWall *)domain_wall_list->Flick();
    bubble_position = ptr->Domain_Position();

    bubble_position_x = bubble_position.Get_Component(0) *
      material_parameters.distance_scale_factor; 
    bubble_position_y = bubble_position.Get_Component(1) *
      material_parameters.distance_scale_factor;

    printf("bubble position: %f %f\n", bubble_position_x, bubble_position_y);


    // the first case is special....
    if(run_count == 0) {

      // if it did not propagate, then search up...
      if(bubble_position_y > ytest  || 
	 bubble_position_x < x_low_test || 
	 bubble_position_x > x_high_test) {
	printf("Bubble did not propagate. Increasing the current.\n");
	lower_current = current;
	current+=current_increment;
	continue;
      }
      // otherwise, it worked, so search down..
      printf("Bubble propagated successfully. Decreasing the current.\n");
      upper_current = current;
      current_increment *= -1.0;
      current += current_increment;
      continue;
    }
    
    // all other cases...

    // searching up, and bubble does not propagate
    // save the limit and increase the current
    if(current_increment > 0.0 && 
       (bubble_position_y > ytest || 
	bubble_position_x < x_low_test || 
	bubble_position_x > x_high_test)) {
      printf("Bubble did not propagate. Increasing the current.\n");
      lower_current = current;
      current+= current_increment;
      continue;
    }

    // searching up and it does propagate
    // save the limit ans quit.
    if(current_increment > 0.0 && 
       (bubble_position_y <= ytest && 
       bubble_position_x >= x_low_test &&
       bubble_position_x <= x_high_test)) {
      printf("Bubble did propagate. Done.\n");
      upper_current = current;
      break;
    }

    // searching down and it does propagate
    // keep going down until it fails
    if(current_increment < 0.0 && 
       bubble_position_y <= ytest && 
       bubble_position_x >= x_low_test && 
       bubble_position_x <= x_high_test) {
      printf("Bubble did propagate. Decreasing the current.\n");
      upper_current = current;
      current+= current_increment;
      continue;
    }

    // searching down and it does not propagate
    // we have found the lower limit.  save it and stop.
    if(current_increment < 0.0 && 
       (bubble_position_y > ytest || 
	 bubble_position_x < x_low_test || 
	 bubble_position_x > x_high_test)) {
      printf("Bubble did not propagate. Done.\n");
      lower_current = current;
      break;
    }
  } // end of the loop....

  if(run_count == max_run_count) {
    printf("Maximum number of iterations has been reached. Aborting.\n");
    return;
  }
  else {
    printf("ML operation bracketed from %f to %f\n",
	   lower_current, upper_current);
  }

  if(lower_current == -1.0 || upper_current == -1.0) {
    printf("Error: one of the limits has not been set.\n");
    return;
  }

  // this is the bracketed search code.
  
  printf("Doing a bracketed search:\n");

  while( fabs(upper_current - lower_current) > tolerance ) {

    current = 0.5 * (lower_current + upper_current);

    sprintf(run_parameters.base_name, "%s-%2.2f", local_base_name, current);
    reset_file_pointers();
    run_parameters.time = 0.0;
    run_parameters.step = 0;
    reset_domain_walls();
    set_all_timing_object_currents(current);  // set the currents

    // do the run...

    printf("Run: name = %s current = %f count = %d\n",
	   run_parameters.base_name, current, run_count);

    status = run_utility(run_time);  

    // check the bubble position....

    domain_wall_list->FReset(); ptr = (DomainWall *)domain_wall_list->Flick();
    bubble_position = ptr->Domain_Position();
    bubble_position_x = bubble_position.Get_Component(0) * material_parameters.distance_scale_factor; 

    bubble_position_y = bubble_position.Get_Component(1) * material_parameters.distance_scale_factor; 

    printf("bubble position: %f %f\n", bubble_position_x, bubble_position_y);

    if(bubble_position_y > ytest  || 
       bubble_position_x < x_low_test || 
       bubble_position_x > x_high_test) {
      printf("bubble did not propagate.\n");
      lower_current = current;
      continue;
    }

    else {
      // in this case, it worked.
      printf("bubble propagation successful\n"); 
      upper_current = current;
      continue;
    }
  }

  printf("ML operation bracketed from %f to %f\n", 
	 lower_current, upper_current);
}



int run_utility(double my_run_length)
{
  DomainWall *ptr;

  // first thing that we do is to recompute the internal parameters...

  calculate_internal_parameters();

  reset_domain_walls();

  // next, we calculate how many steps to take (an integer), and also how often 
  // to write out the state info.

  int nsteps = (int)(my_run_length / input_run_parameters.time_step);
  int summary_steps = (int)(run_parameters.state_save_interval /
			    run_parameters.time_step);
  int dw_n_elements;
  int dw_count;
  int i;

  // loop over time index...
  for(i=0; i<nsteps; i++) {

    // check the number of domain walls, as it may have changed...
    dw_n_elements = domain_wall_list->Number_Of_Elements();
    if( dw_n_elements == 0 ) {
      printf("No domain walls remaining in the list. Time = %f\n", 
	     run_parameters.time * 1.0);
      return -1;
    }

    domain_wall_list->FReset();

    // loop over domain walls.....
    for(dw_count = 0; dw_count < dw_n_elements; dw_count++) {
      ptr = (DomainWall *)domain_wall_list->Flick();
      ptr->Step();
    }

    // update the global time and step count.
    run_parameters.time += run_parameters.time_step;
    run_parameters.step++;

    // code to write bubble/domain wall states out to a file...
    if( i%summary_steps == 0) {
      domain_wall_list->FReset();

      fprintf(run_parameters.state_fp,"state %d %f ", i, run_parameters.time * 
	 material_parameters.time_scale_factor); 

      for(dw_count = 0; dw_count < dw_n_elements; dw_count++) {
	ptr = (DomainWall *)domain_wall_list->Flick();
	ptr->Print_State_And_Phia(run_parameters.state_fp);
      }
      fprintf(run_parameters.state_fp, "\n");
    }


    domain_wall_list->FReset();
    ptr = (DomainWall *)domain_wall_list->Flick();
    Vector max_disp = ptr->Maximum_Wall_Point_Displacement();

    if(max_disp.Get_Component(0) < 1.0e-4 &&
       max_disp.Get_Component(1) < 1.0e-4 &&
       max_disp.Get_Component(2) < 1.0e-4) {
      printf("Wall is at equilibrium. Stopping the run.\n");
      break;
    }
  }

  // Flush all of the buffers and let the user know that the run is done...
  fflush(run_parameters.state_fp);
  fflush(run_parameters.debug_fp);
  fflush(run_parameters.summary_fp);
 
}


void groove_characterization() 
{
  // assume that a bubble and groove have been created, with the 
  // bubble inside the groove. here is the strategy that we will use
  // to extract the collapse field, stripe out field and stripe in
  // field.

  // first make sure that we only have one wall
  if(domain_wall_list->Number_Of_Elements() != 1) {
    printf("Error: There are %d domains. This routine requires 1 domain.\n", 
	   domain_wall_list->Number_Of_Elements());
    return;
  }

  // ok. now get a pointer to the domain...
  DomainWall *ptr;
  domain_wall_list->FReset(); ptr = (DomainWall *)domain_wall_list->Flick();
  
  // and equilibrate it.... here we can limit the time as we are just looking
  // crudely to see if it is going to collapse or stripe out.
  printf("Equilibrating the initial state..\n");
  rte1_status status;
  while((status = run_to_equilibrium_1(300.0)) != equilibrium) {
    if(status == stripe_out) {  // if status is stripe out, 
      ptr->Reset();             // then increase the field and repeat
      input_gbias_field += 10.0;
      continue; 
    }

    if(status == collapse) {   // if the status is collapse,
      ptr->Reset();            // then reduce the field and repeat
      input_gbias_field -= 10.0;
      continue;
    }

    if( status == error) {
      printf("Error: in run_to_equilibrium_1.\n");
      return;
    }
  }
    
  // step up the field by large increment until bubble collapses
  // again, we limit the time here because we are serching for an approximate value

  printf("coarse search up for bubble collapse.....\n");
  input_gbias_field += 5.0;
  while((status = run_to_equilibrium_1(300.0)) != collapse) {
    input_gbias_field += 5.0;
  }
  printf("bubble collapsed at %f (coarse value)\n",input_gbias_field);

  // fine search up for collapse field.
  // go back down to a stable field and equilibrate a bubble.
  // search up until the bubble collapses.

  printf("fine search up for bubble collapse....\n");
  input_gbias_field -= 7.0;
  ptr->Reset(); // reset the bubble
  run_to_equilibrium_1();  // really equilibrate it
  while((status = run_to_equilibrium_1(500.0)) != collapse) {
    input_gbias_field += 1.0;
  }

  double collapse_field = input_gbias_field;
  printf("bubble collapsed at %f Oe (final value)\n", collapse_field);

  // decrease the field by a bit and stabilize a bubble
  input_gbias_field -= 1.0;
  ptr->Reset();  
  run_to_equilibrium_1(300.0);

  // coarse search down for stripe out....
  // again, I specify a limit time here because I am looking for an approximate value...
  while((status = run_to_equilibrium_1(300.0)) != stripe_out) {
    printf("%f\t%f\n", input_gbias_field, ptr->Domain_Radius() * material_parameters.distance_scale_factor);
    input_gbias_field -= 5.0;
  }

  printf("bubble striped out at %f Oe (coarse value)\n", input_gbias_field);

  // reset the bubble and increase the field.  then do a fine
  // search down for the stripe out field.
  // note that I do not specify a limit time here because I suspect that the collapse
  // measurement is sensitive to this...

  input_gbias_field += 7.0;
  ptr->Reset();  
  run_to_equilibrium_1();
  while((status = run_to_equilibrium_1()) != stripe_out) {
    printf("%f\t%f\n", input_gbias_field, ptr->Domain_Radius() * material_parameters.distance_scale_factor);
    input_gbias_field -= 1.0;
  }
  
  double stripe_out_field = input_gbias_field;
  printf("bubble striped out at %f Oe (final value)\n", stripe_out_field);
  
  // ok. now the bubble has started to stripe out.  run it until
  // it is done evolving.

//  while((status = run_to_equilibrium_1()) != equilibrium) {
//    printf("Equilibrating the stripe...\n");
//  }
  printf("done.\n");
  // decrease the field again until it stripes out again.
  // ok. done for now.
}

rte1_status run_to_equilibrium_1(double max_time)
{
  // first thing that we do is to recompute the internal parameters...
  calculate_internal_parameters();

  // next, we calculate how many steps to take (an integer) if specified
  int nsteps = (int)(max_time / input_run_parameters.time_step);

  // get a pointer to the wall and do the run....
  DomainWall *ptr;
  domain_wall_list->FReset(); ptr = (DomainWall *)domain_wall_list->Flick();

  // also, save the initial number of wall points to detect stripe out.
  unsigned inital_number_of_wall_points = ptr->Number_Of_Wall_Points();

  Vector max_disp;

  for(int i=0; ; i++) {
    ptr->Step();
    run_parameters.time += run_parameters.time_step;
    run_parameters.step++;

    // ok. now check for reasons to stop.
    // first look for equilibrium
    max_disp = ptr->Maximum_Wall_Point_Displacement();

    if(max_disp.Get_Component(0) < 1.0e-4 &&
       max_disp.Get_Component(1) < 1.0e-4 &&
       max_disp.Get_Component(2) < 1.0e-4) {
      return equilibrium;
    }

    if(ptr->Number_Of_Wall_Points() > inital_number_of_wall_points + 5) {
      return stripe_out;
    }

    if( ptr->Domain_Radius() * material_parameters.distance_scale_factor 
       < 100.0) {
      return collapse;
    }

    // have we exceeded the maximum time limit ?
    if( nsteps != 0 && i >= nsteps ) break;
  }

  // this is the default. we made it to the end with no exceptions
  return done;
}
