/* Wall-Point.c

   started 1/20/94 by gp


*/

#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "dwm-v1.h"
#include "vector-1.h"
#include "wall-point-v3.h"
#include "dwm-dfield.h"

extern Vector Xhat, Yhat, Zhat;
extern DWMFlickList *field_list;
extern DWMFlickList *groove_field_list;

extern struct MaterialParameters material_parameters;
extern struct RunParameters run_parameters;
extern double gbias_field;
extern double gradient;

DWMfloat Normalize_Angle(DWMfloat angle)
{
  DWMfloat result = angle;

  while( result >= TWOPI ) result -= TWOPI;
  while( result <  0     ) result += TWOPI;

  return(result);
}


DWMfloat Normalize_Angle(DWMfloat angle, DWMfloat reference_angle)
{
  DWMfloat result = angle;

  while( result - reference_angle > PI)  result -= TWOPI;
  while( result - reference_angle < -PI)  result += TWOPI;

  return(result);
}

void WallPoint::initialize_vectors()
{
  // first we allocate space for the vectors.
  Vector temp(3);
  position = temp;
  WallNormal = temp;
  WallTangent = temp;
  Hdemag = temp;
  Hexternal = temp;
  WallForce = temp;
  delta_RSPhi = temp;

  //and now we set some additional values to zero.
  dS_dt = 0.0;

  d_1_plus  = 0.0;
  d_1_0     = 0.0;
  d_1_minus = 0.0;

  d_2_plus  = 0.0;
  d_2_0     = 0.0;
  d_2_minus = 0.0;

  force_r_1 = 0.0;
  force_r_2 = 0.0;
  force_r_3 = 0.0;

}


void WallPoint::Set_Position(Vector new_position, double angle)
{
	position = new_position;
	Phia = angle;
}


void WallPoint::Set_Position(Vector new_position)
{
	position = new_position;
}


void WallPoint::Print_Diagnostics() 
{
	printf("Position = "); position.Print();
	printf("normalized distance = %f\tcurvature = %f\n", 
	       normalized_distance,
	       curvature_1);
	printf("WallNormal = "); WallNormal.Print();
	printf("WallTangent = "); WallTangent.Print();
	printf("Hdemag = ");    Hdemag.Print();
	printf("Hexternal = "); Hexternal.Print();
	printf("WallForce = "); WallForce.Print();
	printf("delta_RSPhi = "); delta_RSPhi.Print();
	printf("Phia = %f\tPhiw = %f\tPhin   = %f\n", 
	       Phia * 57.296, 
	       Phiw * 57.296,
	       Phin * 57.296);
	printf("Next Point is at (norm dist) %f\n", next->normalized_distance);
	printf("\n");

}


void WallPoint::Update_Fields()
{
  double field;

  // x1,y1 is the position in nanometers
  
  double x1 = position.Get_Component(0) * 
    material_parameters.distance_scale_factor;
  double y1 = position.Get_Component(1) * 
    material_parameters.distance_scale_factor;

  field = gbias_field + (gradient * x1);

  double real_time = run_parameters.time * material_parameters.time_scale_factor;

  double destabilizing_field =  0.1 * 
    sin(TWOPI * x1 / 1000.0) * 
      sin(TWOPI * y1 / 1000.0) / material_parameters.four_pi_ms;

  field += destabilizing_field;

  int count = field_list->Number_Of_Elements();
  field_list->FReset();
  int i;

  for(i=0; i<count; i++) {
    field += ((DWMField *)(field_list->Flick()))->Evaluate_Field(x1,y1,real_time) / 
      material_parameters.four_pi_ms;
  }

  count = groove_field_list->Number_Of_Elements();
  groove_field_list->FReset();

  // note that the magnetic fields stored in groove_field_list do not have to be 
  // reduced by 4 pi Ms

  for(i=0; i<count; i++) {
    field += ((DWMField *)(groove_field_list->Flick()))->Evaluate_Field(x1,y1,real_time);
  }

  Hexternal.Set_Component(2, field);
}


void WallPoint::Update_Internals()
{

  delta_s_previous = (position - previous->Position()).Magnitude();
  delta_s_next = (next->Position() - position).Magnitude();

  /* here are parameters that we use to take first and second 
     derivatives along the wall surface */

  double delta_s_total = delta_s_previous + delta_s_next;

  d_1_plus = delta_s_previous / (delta_s_total * delta_s_next);
  d_1_0 = (delta_s_next - delta_s_previous)/(delta_s_next * delta_s_previous);
  d_1_minus = - delta_s_next / (delta_s_total * delta_s_previous);
  
  d_2_plus  = 2.0  / (delta_s_total    * delta_s_next);
  d_2_0     = -2.0 / (delta_s_previous * delta_s_next);
  d_2_minus = 2.0  / (delta_s_total    * delta_s_previous);

  /* Here we calculate the wall tangent and wall normal */

  WallTangent = Position();
  WallTangent.Scale(d_1_0);
  
  Vector Vtemp, Vtemp2;
  Vtemp = next->Position();
  Vtemp.Scale(d_1_plus);
  WallTangent = WallTangent + Vtemp;

  Vtemp = previous->Position();
  Vtemp.Scale(d_1_minus);
  WallTangent = WallTangent + Vtemp;
  WallTangent.Normalize();

  WallNormal  = CrossProduct( WallTangent, Zhat );

  // calculate Phin
  Phin = atan2( DotProduct(WallNormal, Yhat),
	        DotProduct(WallNormal, Xhat) );
  Phin = Normalize_Angle(Phin);

  // calculate Phiw
  Phiw = Phia - Phin;
  Phiw = Normalize_Angle(Phiw);

  // calculate wall curvature

  Vtemp = Position();
  Vtemp.Scale(d_2_0);

  Vtemp2 = next->Position();
  Vtemp2.Scale(d_2_plus);
  Vtemp = Vtemp + Vtemp2;

  Vtemp2 = previous->Position();
  Vtemp2.Scale(d_2_minus);
  Vtemp = Vtemp + Vtemp2;

  curvature_1 = Vtemp.Magnitude();

  double mydot = DotProduct(Vtemp, WallNormal);
  if(mydot >= 0.0 ) curvature_1 *= -1.0;

  // done.
}

void WallPoint::Calculate_Forces()
{
  double prev_Phia, next_Phia;
  double force_r, force_s;

/*****************************
  MOVED
  // calculate wall curvature

  Vtemp = Position();
  Vtemp.Scale(d_2_0);

  Vtemp2 = next->Position();
  Vtemp2.Scale(d_2_plus);
  Vtemp = Vtemp + Vtemp2;

  Vtemp2 = previous->Position();
  Vtemp2.Scale(d_2_minus);
  Vtemp = Vtemp + Vtemp2;

  curvature_1 = Vtemp.Magnitude();

  double mydot = DotProduct(Vtemp, WallNormal);
  if(mydot >= 0.0 ) curvature_1 *= -1.0;
*****************************/


  /* lookout for a branch cut in the variable Phia */

  prev_Phia = previous -> Phia;
  next_Phia = next -> Phia;

  if(Phia - prev_Phia < -1.0)  prev_Phia -= TWOPI;
  if(Phia - prev_Phia >  1.0)  prev_Phia += TWOPI;

  if(Phia - next_Phia < -1.0)  next_Phia -= TWOPI;
  if(Phia - next_Phia >  1.0)  next_Phia += TWOPI;
  
  // calculate the forces acting on the domain wall

  double previous_sin2_Phiw = sin(2.0 * previous->Phiw);
  double next_sin2_Phiw     = sin(2.0 * next->Phiw);

  double dSin2Phiw_dS = 
    d_1_plus *  next_sin2_Phiw +
    d_1_0 * sin(2.0 * Phiw) +
    d_1_minus * previous_sin2_Phiw;


  
    force_r_1 = - material_parameters.quality_factor * curvature_1;
    force_r_2 = - 0.5 * dSin2Phiw_dS;
    force_r_3 = Hdemag.Get_Component(2)     - Hexternal.Get_Component(2);

   force_r =  force_r_1 + force_r_2 + force_r_3;

/*  
   force_r =  
    - material_parameters.quality_factor * curvature_1 
    - 0.5 * dSin2Phiw_dS
    + Hdemag.Get_Component(2) 
    - Hexternal.Get_Component(2);
*/


  double d2Phia_dS2 = 
    d_2_plus  * next_Phia +
    d_2_0     * Phia      +
    d_2_minus * prev_Phia;

/***************************************************
  force_s = 0.5 * sin(2.0 * Phiw) + 
            material_parameters.quality_factor * 
             (next_Phia + prev_Phia - 2.0 * Phia) / (delta_s * delta_s);
**************************************************/

  force_s = 0.5 * sin(2.0 * Phiw) + 
    material_parameters.quality_factor * d2Phia_dS2;
            

  /* Here is were we put in coercivity and wall velocity saturation */
  /* First the coercivity: */

  if( material_parameters.coercivity > 0.0 ) {

    if( fabs(force_r) > material_parameters.coercivity ) {
      if( force_r > 0.0) force_r -= material_parameters.coercivity;
      if( force_r < 0.0) force_r += material_parameters.coercivity;
    }
    else
      force_r = 0.0;
  }

  /* Next we limit the maximum wall velocity */

  if(force_r > material_parameters.maximum_wall_force)
    force_r = material_parameters.maximum_wall_force;
  
  if(force_r < - material_parameters.maximum_wall_force) 
    force_r = - material_parameters.maximum_wall_force;

  WallForce.Set_Component(0, force_r);
  WallForce.Set_Component(1, force_s);
}

void WallPoint::Calculate_Delta_Phia()
{
  DWMfloat prev_Phia = previous -> Phia;
  DWMfloat next_Phia = next -> Phia;

  if(Phia - prev_Phia < -1.0)  prev_Phia -=TWOPI;
  if(Phia - prev_Phia >  1.0)  prev_Phia +=TWOPI;

  if(Phia - next_Phia < -1.0)  next_Phia -=TWOPI;
  if(Phia - next_Phia >  1.0)  next_Phia +=TWOPI;

/**********************************************************************
  double delta_s_total =delta_s_previous + delta_s_next;
  double dPhia_dS =
    (delta_s_previous / (delta_s_total * delta_s_next)) * next_Phia -
    (delta_s_next / (delta_s_total * delta_s_previous)) * prev_Phia + 
    ((delta_s_next - delta_s_previous) / 
     (delta_s_next * delta_s_previous)) * Phia;
**********************************************************************/

  double dPhia_dS = 
    d_1_minus * prev_Phia +
    d_1_0     * Phia +
    d_1_plus  * next_Phia;

  double foo = delta_RSPhi.Get_Component(2) + dPhia_dS * delta_RSPhi.Get_Component(1);

  delta_RSPhi.Set_Component(2,foo);
}

void WallPoint::Update_Position()
{
  /* translate delta_RSPhi into rectangular coordinates and update the wall
     position and magnetization */

  Vector displacement_r, displacement_s;

  displacement_r = WallNormal;
  displacement_s = WallTangent;
  
  displacement_r.Scale(delta_RSPhi.Get_Component(0));
  displacement_s.Scale(delta_RSPhi.Get_Component(1));

  position = position + displacement_r;
  position = position + displacement_s;

  Phia += delta_RSPhi.Get_Component(2);
  Phia = Normalize_Angle(Phia);
}


